#ifndef __EBR_H__
#define __EBR_H__

#include <stdio.h>
#include <HECH.h>
#include <Properties.h>
#include <Battery.h>
#include <CommandLineInterface.h>


#define PACKAGE_STRING "EBR Control Software"
#define VERSION_STRING "2.0"
#define COPYRIGHT_STRING "(C) 2014\nMCLab Group at the Department of Computer Science\nSapienza University of Rome\n"


/* Linux defines PATH_MAX in a non standard library, so we redefine it! */
#define PATH_MAX        4096

/* Possibly defined in the Environment */
#ifndef TIMESLOTS_AHEAD
#define TIMESLOTS_AHEAD	6
#endif

#define TIMESLOTS_IN_A_DAY 24

#define TOLERABLE_PHEV_ABSENCE  0.6

#define PHEV_CHARGE		0.7
#define PHEV_READY_HOURS_MIN	4


#define TOOL_DEFAULT	"EBR-CTRL"
#define TARIFF_DEFAULT	"IBR"


#define MAX_BUF_LEN (1 << 20)


#define PAST_DISTRIBUTION "0.5:0.25:0.125:0.0625:0.03125:0.015625:0.0078125:0.00390625:0.00195313:0.00195312"



typedef enum tool_t {
	TOOL_CTRL = 0,
	TOOL_EVAL,
	TOOL_HISTORY,
	TOOL_SUBST_CTRL,
	TOOL_SUBST_HISTORY,
	TOOL_NUMBER
} Tool_t;

typedef enum forecasts_t {
	FORECASTS_CONSUMPTION = 0,
	FORECASTS_PRODUCTION,
	FORECASTS_PHEV,
	FORECASTS_NUMBER
} Forecasts_t;

typedef enum pricepolicy_t {
	PRICEPOLICY_MIN = 0,
	PRICEPOLICY_MAX,
	PRICEPOLICY_NUMBER
} PricePolicy_t;

typedef enum tariff_t {
	TARIFF_TOU = 0,
	TARIFF_IBR,
	TARIFF_FLAT,
	TARIFF_SUBST,
	TARIFF_NUMBER
} Tariff_t;

typedef enum tou_t {
	TOU_BUY = 0,
	TOU_SELL,
	TOU_NUMBER
} ToU_t;

typedef enum ibr_t {
	IBR_BUY_LOW = 0,
	IBR_BUY_HIGH,
	IBR_SELL,
	IBR_NUMBER
} IBR_t;



typedef struct EBR {

	/* Control strategy to use */
	Tool_t tool;

	/* tariffs EUR/kWh */
	Tariff_t tariff;
	double ToU[24][TOU_NUMBER];
	int currentHour; /* needed for applying correct ToU */
	double IBR[IBR_NUMBER];
	double FLAT[TOU_NUMBER];

	/* Price Policy for IBR */
	char pricePolicyFile[PATH_MAX];
	FILE * pricePolicyFile_fd;
	unsigned long pricePolicyFileCurrentLine;
	/* price policy in the future this->timeslotsAhead slots, for MILP */
	double * pricePolicy[PRICEPOLICY_NUMBER];

	/* Energy cost for SUBST tariff */
	char tariffsSubstFile[PATH_MAX];
	FILE * tariffsSubstFile_fd;
	double SUBSToU[TIMESLOTS_AHEAD];
	int tariffsSubstFile_started;

	/* Past management for EBR-CTRL (for forecasting future) */
	int maxPastLength;
	double * pastDistribution;
	/* one cursor pastLength for each timeslot */
	int pastLength[TIMESLOTS_IN_A_DAY];
	/* one array past for each timeslot and for each type */
	double * past[FORECASTS_NUMBER][TIMESLOTS_IN_A_DAY];

	/* 
	 *	Use real historical data:
	 *	- to forecast next TIMESLOTS_AHEAD timeslots (EBR-Eval) 
	 *	- to simulate history (EBR-HISTORY)
	 */
	char useHistoryFile[PATH_MAX];
	int useHistory;
	FILE * useHistoryFile_fd;

	/* Energy Window for TOOL_SUBST_CTRL */
	int energyWindowSubst_l;
	double * energyWindowSubst;
	int energyWindowSubst_index;

	const char * configFile;
	Properties * configHECH;

	char loggingFile[PATH_MAX];
	FILE *loggingFile_fd;

	char batteryFile[PATH_MAX];
	Battery * battery;

	char phevFile[PATH_MAX];
	Battery * PHEV;

	/* to be taken in --energyFile */
	double maxEnergy;
	double minEnergy;

	/* time slot length in minutes */
	int timeslotLength;
	int timeslotsInHour;

	/* number of hours ahead to consider for forecasting */
	double forecastHours;
	int timeslotsAhead;

	/* used to manage training of forecast */
	int trainingForecast;

} EBR;


EBR * EBR_new(CommandLineInterface * cliPtr);
void EBR_free(EBR ** _this);

void EBR_readToUTariffs(EBR *_this, const char * tariffsToUFile);

void EBR_updateCurrentHour(EBR *_this);


/**
 *	SUBST ToU tariff
 */
int EBR_stepSubstTariff(EBR *_this);

void EBR_dumpSubstTariff(EBR *_this);

void EBR_dumpEnergyWindow(EBR *_this);


/**
 *	Stores current (and future) time-slot price policies into parameter _this->pricePolicy
 *	Returns 0 is everything is fine, -1 otherwise
 */
int EBR_stepPricePolicy(EBR *_this);

void EBR_dumpPricePolicy(EBR *_this);


/**
 *	Utilities
 */
void EBR_printBanner(FILE *);

CommandLineInterface * EBR_CLI(int argc, char ** argv);

void EBR_logMeasurements(EBR *_this, double * measurements, double * commands, double forecast[FORECASTS_NUMBER][TIMESLOTS_AHEAD], double energyForecast, double gridEnergyCost);

void EBR_printForecasts(int timeslot, double forecast[FORECASTS_NUMBER][TIMESLOTS_AHEAD]);

double EBR_energyFromGrid(EBR *_this, double * measurements, double * commands);



/**
 *	Prices, returns applied tariff
 */
double EBR_energyPrice(EBR *_this, double energy, double * energyPrice);




/**
 *	HECH Interface
 */
HECH * EBR_houseConnect(Properties *conf);
void EBR_houseDisconnect(HECH **);




/**
 *	Decides commands to give to battery and phev, if present
 *	@parameter commands must be an array of at least size 2
 */
int EBR_decideCommands(EBR *_this, int timeslot, double forecast[FORECASTS_NUMBER][TIMESLOTS_AHEAD], double *measurements, double *commands, double *gridEnergy, double * gridEnergyCost);

#if 0
int EBR_decideCommands_greedy(EBR *_this, int timeslot, double forecast[FORECASTS_NUMBER][TIMESLOTS_AHEAD], double *measurements, double *commands, double *gridEnergy);
#endif

int EBR_decideCommands_optimization(EBR *_this, int timeslot, double forecast[FORECASTS_NUMBER][TIMESLOTS_AHEAD], double *measurements, double *commands, double *gridEnergy, double * gridEnergyCost);


char * EBR_create_milp_for_control(EBR *_this, int timeslot, double forecast[FORECASTS_NUMBER][TIMESLOTS_AHEAD], double * measurements);

char * EBR_create_milp_for_control_subst(EBR *_this, int timeslot, double forecast[FORECASTS_NUMBER][TIMESLOTS_AHEAD], double * measurements);



/**
 *	Function to compute forecast for the TIMESLOTS_AHEAD hours ahead, basing on the past
 *	First position of results refers to timeslot+1, last one to timeslot+TIMESLOTS_AHEAD
 *	Pre: double * parameters must be allocated with size TIMESLOTS_AHEAD
 *	If everything is OK it properly fills parameters.
 *	If in simulation mode, commands[CMDS_PHEV] is properly filled according to history
 */
int EBR_forecast(EBR *_this, int timeslot, double *consumption, double *production, double *PHEV);


void EBR_readPHEVCommand(EBR *_this, double * commands);


void EBR_substEnergyWindowPush(EBR * _this, double energy);



/**
 *	Learn from past
 */
void EBR_pastPush(EBR *_this, int timeslot, double consumption, double production, double PHEV);

void EBR_pastDump(EBR *_this, int timeslot);


int EBR_isPastTraining(EBR *_this);

void EBR_initPastTraining(EBR *_this);

void EBR_nextPastTraining(EBR *_this);


#endif

