#include <EUMF-K.h>
#include <EUMF-K_milp.h>
#include <Debug.h>
#include <Battery.h>

#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include <strings.h>


#define DEBUG "EBR"


int EBR_decideCommands(EBR *_this, int timeslot, double forecast[FORECASTS_NUMBER][TIMESLOTS_AHEAD], double *measurements, double *commands, double *gridEnergy, double * gridEnergyCost)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	Debug_assert(DEBUG, _this->tool == TOOL_CTRL || _this->tool == TOOL_EVAL || _this->tool == TOOL_SUBST_CTRL, "!(_this->tool == TOOL_CTRL || _this->tool == TOOL_EVAL || _this->tool == TOOL_SUBST_CTRL)\n");

	int res;
	res = EBR_decideCommands_optimization(_this, timeslot, forecast, measurements, commands, gridEnergy, gridEnergyCost);
	return res;
}



int EBR_decideCommands_optimization(EBR *_this, int timeslot, double forecast[FORECASTS_NUMBER][TIMESLOTS_AHEAD], double *measurements, double *commands, double * gridEnergy, double * gridEnergyCost)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	Debug_assert(DEBUG, _this->tool == TOOL_CTRL || _this->tool == TOOL_EVAL || _this->tool == TOOL_SUBST_CTRL, "!(_this->tool == TOOL_CTRL || _this->tool == TOOL_EVAL || _this->tool == TOOL_SUBST_CTRL)\n");

	static char sql[MAX_BUF_LEN];
	static char solver[100];
	static char filename_log[MAX_BUF_LEN];

	char * filename;
	MILPSolver_t res;
	int answer;

	/* Create MILP file */
	if (_this->tool == TOOL_CTRL || _this->tool == TOOL_EVAL) {
		filename = EBR_create_milp_for_control(_this, timeslot, forecast, measurements);
	}
	else {
		filename = EBR_create_milp_for_control_subst(_this, timeslot, forecast, measurements);
	}
	if (!filename) {
		fprintf(stderr, "No MILP created\n");
		return -1;
	}
	Debug_out(DEBUG, "MILP '%s/%s.lp' created\n", MILP_DIR, filename);

	/* Solve MILP file */
	do {
		res = EBR_milp_solve_file(filename, MILP_DIR, sql, filename_log, MAX_BUF_LEN);
		if (res == MILPSOLVER_NONE) {
			fprintf(stderr, "Neither cplex nor glpsol found, sleeping %u seconds\n", WAIT_SOLUTION);
			sleep(WAIT_SOLUTION);
		}
	} while (res == MILPSOLVER_NONE);
	EBR_milp_solver(res, solver);
	Debug_out(DEBUG, "MILP solved with %s, solution in file '%s/%s'\n", solver, MILP_DIR, sql);

	/* Parse result */
	answer = EBR_milp_get_solution(res, MILP_DIR, sql, filename_log, commands, gridEnergy, gridEnergyCost);
	if (answer < 0) {
		fprintf(stderr, "No solution to MILP '%s/%s.lp' at timeslot %d\n", MILP_DIR, filename, timeslot);
		return -1;
	}

	free(filename);
	filename = NULL;
	
	return 0;
}



/**
 *	MILP for House
 *	Return MILP file name on success, NULL otherwise
 */

char * EBR_create_milp_for_control(EBR *_this, int timeslot, double forecast[FORECASTS_NUMBER][TIMESLOTS_AHEAD], double * measurements)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	Debug_assert(DEBUG, _this->tool == TOOL_CTRL || _this->tool == TOOL_EVAL, "!(_this->tool == TOOL_CTRL || _this->tool == TOOL_EVAL)\n");

	static char var_name[1024];
	static char * varnames_for_row[TIMESLOTS_AHEAD];
	static char obj_function[20*TIMESLOTS_AHEAD];
	static char sql[MAX_BUF_LEN];
	static double vals_for_row[TIMESLOTS_AHEAD];

	double d_minus_r;
	double phevCapacity;

	char * filename = NULL;
	double maxBattery, minBattery, maxCharge, maxDischarge;
	int ts_len, ts_len_tmp;
	int t, i;
	time_t timestamp_time;
	double timestamp;
	int isPHEV[TIMESLOTS_AHEAD];
	int phevHours = 0;
	int howmany;
	int isThereBattery;
	int isTherePHEV;

	isThereBattery = Battery_getCapacity(_this->battery);
	// PHEV could be absent, or plugged with 0 capacity (which is equivalent)
	isTherePHEV = _this->PHEV && Battery_getCapacity(_this->PHEV);

	pid_t myPid = getpid();

	FILE * stream_vars_bnds;
	FILE * stream_vars_list;
	FILE * stream_conv;

	time(&timestamp_time);
	timestamp = difftime(timestamp_time, 0);

	/* Adjust forecast for PHEV according to PHEV_READY_HOURS, if any */
	if (isTherePHEV && measurements[MEAS_PHEV] > 0 && measurements[MEAS_PHEV_READY_HOURS] > 0) {
		phevHours = 1; /* 1 because it counts also present time, which is timeslot t = 1 */
		for (t = 0; t < TIMESLOTS_AHEAD-1 && t < measurements[MEAS_PHEV_READY_HOURS]-1; t++) {
			forecast[FORECASTS_PHEV][t] = measurements[MEAS_PHEV];
			phevHours++;
		}
		forecast[FORECASTS_PHEV][t] = -1;
	}
//	Debug_out(DEBUG, "AFTER REARRANGING FORECAST\n");
//	EBR_printForecasts(timeslot, forecast);

	/* stores when PHEV is present */
	if (isTherePHEV) {
		for (t = 0; t < TIMESLOTS_AHEAD; t++) {
			isPHEV[t] = (t == 0 && measurements[MEAS_PHEV] == -1) || (t > 0 && forecast[FORECASTS_PHEV][t-1] == -1)? 0 : 1;
		}
	}
	else {
		bzero(isPHEV, TIMESLOTS_AHEAD);
	}

	/* STREAMS */
	for (ts_len = 0, ts_len_tmp = timeslot;
		 ts_len_tmp > 0;
		 ts_len++, ts_len_tmp = ts_len_tmp/10);
	filename = (char *) calloc(strlen(MILP_DIR) + ts_len + 100, sizeof(char));
	sprintf(filename, "%s/%d.%.0f.%d.lp.bnds", MILP_DIR, myPid, timestamp, timeslot);
	if (!(stream_vars_bnds = fopen(filename, "w"))) {
		fprintf(stderr, "Impossible to open file %s: ", filename);
		perror("");
		exit(1);
	}
	sprintf(filename, "%s/%d.%.0f.%d.lp.bool", MILP_DIR, myPid, timestamp, timeslot);
	if (!(stream_vars_list = fopen(filename, "w"))) {
		fprintf(stderr, "Impossible to open file %s: ", filename);
		perror("");
		exit(1);
	}
	sprintf(filename, "%s/%d.%.0f.%d.clp", MILP_DIR, myPid, timestamp, timeslot);
	if (!(stream_conv = fopen(filename, "w"))) {
		fprintf(stderr, "Impossible to open file %s: ", filename);
		perror("");
		exit(1);
	}

	/* VARIABLES */

	/**
	 *	For all energy storage appliances i, for all timeslots in the future
	 *	- b_i(t) in [20% B_iu, B_iu]
	 *		- b_i(1) fixed as measured
	 *	- a+_i(t) in [0, C_iu]
	 *	- a-_i(t) in [0, C_il]
	 *	- y_it	in B
	 */
	/* Battery = 1 */
	maxBattery = Battery_getCapacity(_this->battery);
	if (isThereBattery) {
		minBattery = 20.0*maxBattery/100.0;
		maxCharge = Battery_getChargeSpeed(_this->battery);
		maxDischarge = Battery_getDischargeSpeed(_this->battery);
		for (t = 1; t <= TIMESLOTS_AHEAD; t++) {
			if (t > 1) {
				/* b_i(t) */
				sprintf(var_name, "b_1_%d", t);
				EBR_create_var_for_milp(stream_conv, (FILE *)0, stream_vars_bnds, var_name, minBattery, maxBattery);
			}
			/* a+_i(t) charge */
			sprintf(var_name, "a_1_c_%d", t);
			EBR_create_var_for_milp(stream_conv, (FILE *)0, stream_vars_bnds, var_name, 0, maxCharge);
			/* a-_i(t) discharge */
			sprintf(var_name, "a_1_d_%d", t);
			EBR_create_var_for_milp(stream_conv, (FILE *)0, stream_vars_bnds, var_name, 0, maxDischarge);
			/* y_it */
			sprintf(var_name, "y_1_%d", t);
			EBR_create_var_for_milp(stream_conv, stream_vars_list, stream_vars_bnds, var_name, 0, 1);
		}
		/* Add also last TIMESLOTS_AHEAD+1 battery state of charge variable */
		sprintf(var_name, "b_1_%d", TIMESLOTS_AHEAD+1);
		EBR_create_var_for_milp(stream_conv, (FILE *)0, stream_vars_bnds, var_name, minBattery, maxBattery);
	}

	/* PHEV = 2 (PHEV does not discharge!) */
	if (isTherePHEV) {
		maxBattery = Battery_getCapacity(_this->PHEV);
		minBattery = 20.0*maxBattery/100.0;
		maxCharge = Battery_getChargeSpeed(_this->PHEV);
		for (t = 1; t <= TIMESLOTS_AHEAD; t++) {
			if (t > 1 && isPHEV[t-2]) {
				/* b_i(t) */
				sprintf(var_name, "b_2_%d", t);
				EBR_create_var_for_milp(stream_conv, (FILE *)0, stream_vars_bnds, var_name, minBattery, maxBattery);
			}
			if (isPHEV[t-1]) {
				/* a+_i(t) charge */
				sprintf(var_name, "a_2_c_%d", t);
				EBR_create_var_for_milp(stream_conv, (FILE *)0, stream_vars_bnds, var_name, 0, maxCharge);
			}
		}
		if (isPHEV[TIMESLOTS_AHEAD-1]) {
			/* b_i(t) */
			sprintf(var_name, "b_2_%d", TIMESLOTS_AHEAD+1);
			EBR_create_var_for_milp(stream_conv, (FILE *)0, stream_vars_bnds, var_name, minBattery, maxBattery);
		}
	}

	/**
	 *	Energy from (to) grid
	 *	- e(t)	in [COMPUTE BOUNDS]
	 */
	for (t = 1; t <= TIMESLOTS_AHEAD; t++) {
		sprintf(var_name, "e_%d", t);
		EBR_create_var_for_milp(stream_conv, (FILE *)0, stream_vars_bnds, var_name, _this->minEnergy, _this->maxEnergy);
	}

	/**
	 *	Cost function
	 *	- Gamma(t)	in [COMPUTE BOUNDS]
	 *	- d_t0, d_t1, d_t2, d_t3	in B
	 */
	if (_this->tariff == TARIFF_TOU) {
		/* Gamma_t */
		double minCost = 0, maxCost = 0;
		for (i = 0; i < 24; i++) {
			maxCost = maxCost < _this->ToU[i][TOU_BUY]? _this->ToU[i][TOU_BUY] : maxCost;
			minCost = minCost < _this->ToU[i][TOU_SELL]? _this->ToU[i][TOU_SELL] : minCost;
		}
		for (t = 1; t <= TIMESLOTS_AHEAD; t++) {
			sprintf(var_name, "Gamma_%d", t);
			EBR_create_var_for_milp(stream_conv, (FILE *)0,
									stream_vars_bnds, var_name,
									minCost * (_this->minEnergy),
									maxCost *(_this->maxEnergy));
			/* auxiliary variable */
			sprintf(var_name, "d_%d", t);
			EBR_create_var_for_milp(stream_conv, stream_vars_list, stream_vars_bnds, var_name, 0, 1);
		}
	}
	else { /* _this->tariff == TARIFF_IBR */
		for (t = 1; t <= TIMESLOTS_AHEAD; t++) {
			/* Gamma_t */
			sprintf(var_name, "Gamma_%d", t);
			EBR_create_var_for_milp(stream_conv, (FILE *)0,
									stream_vars_bnds, var_name,
									(_this->IBR[IBR_SELL])*(_this->minEnergy),
									(_this->IBR[IBR_BUY_HIGH])*(_this->maxEnergy));
			/* auxiliary variables */
			sprintf(var_name, "d_%d_0", t);
			EBR_create_var_for_milp(stream_conv, stream_vars_list, stream_vars_bnds, var_name, 0, 1);
			sprintf(var_name, "d_%d_1", t);
			EBR_create_var_for_milp(stream_conv, stream_vars_list, stream_vars_bnds, var_name, 0, 1);
			sprintf(var_name, "d_%d_2", t);
			EBR_create_var_for_milp(stream_conv, stream_vars_list, stream_vars_bnds, var_name, 0, 1);
			sprintf(var_name, "d_%d_3", t);
			EBR_create_var_for_milp(stream_conv, stream_vars_list, stream_vars_bnds, var_name, 0, 1);
		}
	}

	/* CONSTRAINTS */

	/* Battery */
	if (isThereBattery) {
		maxCharge = Battery_getChargeFactor(_this->battery);
		/* b_i(2) = b_i(1) + T/60 ( alpha_i * a+_i(1) - a-_i(1) ) */
		/* b_i(2) - T/60 * alpha_i * a+_i(1) + T/60 * a-_i(1) = b_i(1) (= const) */
		EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "b_1_2");
		EBR_milp_monomial(varnames_for_row, vals_for_row, 1, -maxCharge, "a_1_c_1");
		EBR_milp_monomial(varnames_for_row, vals_for_row, 2, 1, "a_1_d_1");
		EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_FIXED, measurements[MEAS_BATTERY], measurements[MEAS_BATTERY], NULL, 0, 3, varnames_for_row, vals_for_row);
		for (t = 1; t <= TIMESLOTS_AHEAD; t++) {
			/* b_i(t+1) = b_i(t) + T/60 ( alpha_i * a+_i(t) - a-_i(t) ) */
			/* b_i(t+1) - b_i(t) - T/60 * alpha_i * a+_i(t) + T/60 * a-_i(t) = 0 */
			if (t > 1) {
				EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "b_1_%d", t+1);
				EBR_milp_monomial(varnames_for_row, vals_for_row, 1, -1, "b_1_%d", t);
				EBR_milp_monomial(varnames_for_row, vals_for_row, 2, -maxCharge, "a_1_c_%d", t);
				EBR_milp_monomial(varnames_for_row, vals_for_row, 3, 1, "a_1_d_%d", t);
				EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_FIXED, 0, 0, NULL, 0, 4, varnames_for_row, vals_for_row);
			}

			/*     y_it -> a+_i(t) = 0 */
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "a_1_c_%d", t);
			sprintf(var_name, "y_1_%d", t);
			EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_FIXED, 0, 0, var_name, 0, 1, varnames_for_row, vals_for_row);

			/* not y_it -> a-_i(t) = 0 */
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "a_1_d_%d", t);
			sprintf(var_name, "y_1_%d", t);
			EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_FIXED, 0, 0, var_name, 1, 1, varnames_for_row, vals_for_row);
		}
	}

	/* PHEV (does not discharge) */
	if (isTherePHEV) {
		maxCharge = Battery_getChargeFactor(_this->PHEV);
		for (t = 1; t <= TIMESLOTS_AHEAD; t++) {
			/* b_i(t+1) = b_i(t) + T/60 alpha_i * a+_i(t) */
			/* b_i(t+1) - b_i(t) - T/60 * alpha_i * a+_i(t) = 0 */
			if (isPHEV[t-1]) {
				/* b_i(t+1) = b_i(t) + T/60 alpha_i * a+_i(t) */
				if (t == 1 || !isPHEV[t-2]) {
					/* phev will plug in timeslot t */
					EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "b_2_%d", t+1);
					EBR_milp_monomial(varnames_for_row, vals_for_row, 1, -maxCharge, "a_2_c_%d", t);
					if (t == 1) {
						EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_FIXED, measurements[MEAS_PHEV], measurements[MEAS_PHEV], NULL, 0, 2, varnames_for_row, vals_for_row);
					}
					else {
						EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_FIXED, forecast[FORECASTS_PHEV][t-2], forecast[FORECASTS_PHEV][t-2], NULL, 0, 2, varnames_for_row, vals_for_row);
					}
				}
				else {
					/* phev will be already plugged in timeslot t */
					EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "b_2_%d", t+1);
					EBR_milp_monomial(varnames_for_row, vals_for_row, 1, -1, "b_2_%d", t);
					EBR_milp_monomial(varnames_for_row, vals_for_row, 2, -maxCharge, "a_2_c_%d", t);
					EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_FIXED, 0, 0, NULL, 0, 3, varnames_for_row, vals_for_row);
				}
				/* b_i(t+1) >= PHEV_CHARGE * capacity_i(t) */
				if (t == TIMESLOTS_AHEAD || !isPHEV[t]) {
#define TIMESLOTS_PAST_MIN 4
					/* exclude the following case: t == TIMESLOTS_AHEAD and phev was plugged less than TIMESLOTS_PAST_MIN time slots earlier */
					if (t == TIMESLOTS_AHEAD && isPHEV[t - 1]) {
						int t_past, continue_for = 0;
						for (t_past = t - 2; t_past >= t - TIMESLOTS_PAST_MIN; t_past--) {
							Debug_out(DEBUG, "To avoid infeasibility, checking time-slot %d out of %d (%d) --> %d\n", t_past, t, isPHEV[t_past], continue_for);
							if (!isPHEV[t_past]) {
								continue_for = 1;
								break;
							}
						}
						if (continue_for) {
							Debug_out(DEBUG, "Continuing!\n");
							continue;
						}
					}

					if (phevHours > 0) {
						static double computes[2];
						computes[0] = measurements[MEAS_PHEV] + ((double)phevHours/measurements[MEAS_PHEV_READY_HOURS]) * maxCharge * Battery_getChargeSpeed(_this->PHEV);
						computes[1] = ((double)phevHours/measurements[MEAS_PHEV_READY_HOURS])*Battery_getCapacity(_this->PHEV);
						phevCapacity = computes[0] < computes[1]? computes[0] : computes[1];
						phevHours = 0;
					}
					else {
						if (t == TIMESLOTS_AHEAD && isPHEV[t]) {
							for (i = t-1, phevHours = 0; i >= 0 && isPHEV[i]; i--, phevHours++);
							phevCapacity = ((double)phevHours/(double)TIMESLOTS_AHEAD) * Battery_getCapacity(_this->PHEV);
						}
						else {
							phevCapacity = PHEV_CHARGE * Battery_getCapacity(_this->PHEV);
						}
					}
					EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "b_2_%d", t+1);
					EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_LB, phevCapacity, phevCapacity, NULL, 0, 1, varnames_for_row, vals_for_row);
				}
			}
		}
	}

	/* Energy balancing */
	/* e(t) + (beta_1 * a-_1(t)) + r(t) = d(t) + sum_{i in 1,2}(a+_i(t)) */
	/* e(t) + (beta_1 * a-_1(t)) - sum_{i in 1,2}(a+_i(t)) = d(t) - r(t) */
	/* maxDischarge = discharge factor of battery */
	maxDischarge = Battery_getDischargeFactor(_this->battery);
	for (t = 1; t <= TIMESLOTS_AHEAD; t++) {
		howmany = 0;
		if (t == 1) {
			d_minus_r = measurements[MEAS_CONSUMPTION] - measurements[MEAS_PRODUCTION];
		}
		else {
			d_minus_r = forecast[FORECASTS_CONSUMPTION][t-2] - forecast[FORECASTS_PRODUCTION][t-2];
		}
		EBR_milp_monomial(varnames_for_row, vals_for_row, howmany++, 1, "e_%d", t);
		if (isThereBattery) {
			EBR_milp_monomial(varnames_for_row, vals_for_row, howmany++, maxDischarge, "a_1_d_%d", t);
			EBR_milp_monomial(varnames_for_row, vals_for_row, howmany++, -1, "a_1_c_%d", t);
		}
		if (isTherePHEV && isPHEV[t-1]) {
			EBR_milp_monomial(varnames_for_row, vals_for_row, howmany++, -1, "a_2_c_%d", t);
		}
		EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_FIXED, d_minus_r, d_minus_r, NULL, 0, howmany, varnames_for_row, vals_for_row);
	}

	/* Cost */
	if (_this->tariff == TARIFF_TOU) {
		for (t = 1; t <= TIMESLOTS_AHEAD; t++) {
			/* choose right ToU tariff */
			int tou_t = (timeslot-1) + t - 1;
			if (tou_t >= TIMESLOTS_IN_A_DAY) {
				tou_t -= TIMESLOTS_IN_A_DAY;
			}
			// d_t -> e(t) <= 0
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "e_%d", t);
			sprintf(var_name, "d_%d", t);
			EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_UB, 0, 0, var_name, 0, 1, varnames_for_row, vals_for_row);

			// !d_t -> e(t) >= 0
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "e_%d", t);
			sprintf(var_name, "d_%d", t);
			EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_LB, 0, 0, var_name, 1, 1, varnames_for_row, vals_for_row);

			// d_t -> Γ(t) = (T/60) s(t) e(t)
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "Gamma_%d", t);
			EBR_milp_monomial(varnames_for_row, vals_for_row, 1, -(_this->ToU[tou_t][TOU_SELL]), "e_%d", t);
			sprintf(var_name, "d_%d", t);
			EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_FIXED, 0, 0, var_name, 0, 2, varnames_for_row, vals_for_row);
			Debug_out(DEBUG, "Added constraint: d_%d -> Gamma_%d + %lg e_%d = 0 (tou_t = %d)\n", t, t, -(_this->ToU[tou_t][TOU_SELL]), t, tou_t);

			// !d_t -> Γ(t) = (T/60) c(t) e(t)
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "Gamma_%d", t);
			EBR_milp_monomial(varnames_for_row, vals_for_row, 1, -(_this->ToU[tou_t][TOU_BUY]), "e_%d", t);
			sprintf(var_name, "d_%d", t);
			EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_FIXED, 0, 0, var_name, 1, 2, varnames_for_row, vals_for_row);
			Debug_out(DEBUG, "Added constraint: !d_%d -> Gamma_%d + %lg e_%d = 0 (tou_t = %d)\n", t, t, -(_this->ToU[tou_t][TOU_BUY]), t, tou_t);
		}
	}
	else { /* _this->tariff == TARIFF_IBR */
		for (t = 1; t <= TIMESLOTS_AHEAD; t++) {
			// d_t_0 -> e(t) <= 0
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "e_%d", t);
			sprintf(var_name, "d_%d_0", t);
			EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_UB, 0, 0, var_name, 0, 1, varnames_for_row, vals_for_row);

			// d_t_1 -> e(t) <= gamma-(t)
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "e_%d", t);
			sprintf(var_name, "d_%d_1", t);
			EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_UB, 0, _this->pricePolicy[PRICEPOLICY_MIN][t-1], var_name, 0, 1, varnames_for_row, vals_for_row);

			// d_t_1 -> e(t) >= 0
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "e_%d", t);
			sprintf(var_name, "d_%d_1", t);
			EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_LB, 0, 0, var_name, 0, 1, varnames_for_row, vals_for_row);

			// d_t_2 -> e(t) >= gamma+(t)
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "e_%d", t);
			sprintf(var_name, "d_%d_2", t);
			EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_LB, _this->pricePolicy[PRICEPOLICY_MAX][t-1], 0, var_name, 0, 1, varnames_for_row, vals_for_row);

			// d_t_2 -> e(t) >= 0
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "e_%d", t);
			sprintf(var_name, "d_%d_2", t);
			EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_LB, 0, 0, var_name, 0, 1, varnames_for_row, vals_for_row);

			// d_t_3 -> e(t) >= gamma-(t)
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "e_%d", t);
			sprintf(var_name, "d_%d_3", t);
			EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_LB, _this->pricePolicy[PRICEPOLICY_MIN][t-1], 0, var_name, 0, 1, varnames_for_row, vals_for_row);

			// d_t_3 -> e(t) <= gamma+(t)
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "e_%d", t);
			sprintf(var_name, "d_%d_3", t);
			EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_UB, 0, _this->pricePolicy[PRICEPOLICY_MAX][t-1], var_name, 0, 1, varnames_for_row, vals_for_row);

			// d_t_3 -> e(t) >= 0
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "e_%d", t);
			sprintf(var_name, "d_%d_3", t);
			EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_LB, 0, 0, var_name, 0, 1, varnames_for_row, vals_for_row);

			// d_t_0 + d_t_1 + d_t_2 + d_t_3 = 1
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "d_%d_0", t);
			EBR_milp_monomial(varnames_for_row, vals_for_row, 1, 1, "d_%d_1", t);
			EBR_milp_monomial(varnames_for_row, vals_for_row, 2, 1, "d_%d_2", t);
			EBR_milp_monomial(varnames_for_row, vals_for_row, 3, 1, "d_%d_3", t);
			EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_FIXED, 1, 1, NULL, 0, 4, varnames_for_row, vals_for_row);

			// d_t_0 -> Γ(t) = (T/60) s(t) e(t)
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "Gamma_%d", t);
			EBR_milp_monomial(varnames_for_row, vals_for_row, 1, -(_this->IBR[IBR_SELL]), "e_%d", t);
			sprintf(var_name, "d_%d_0", t);
			EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_FIXED, 0, 0, var_name, 0, 2, varnames_for_row, vals_for_row);

			// d_t_1 -> Γ(t) = (T/60) hb(t) e(t)
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "Gamma_%d", t);
			EBR_milp_monomial(varnames_for_row, vals_for_row, 1, -(_this->IBR[IBR_BUY_HIGH]), "e_%d", t);
			sprintf(var_name, "d_%d_1", t);
			EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_FIXED, 0, 0, var_name, 0, 2, varnames_for_row, vals_for_row);

			// d_t_2 -> Γ(t) = (T/60) hb(t) e(t)
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "Gamma_%d", t);
			EBR_milp_monomial(varnames_for_row, vals_for_row, 1, -(_this->IBR[IBR_BUY_HIGH]), "e_%d", t);
			sprintf(var_name, "d_%d_2", t);
			EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_FIXED, 0, 0, var_name, 0, 2, varnames_for_row, vals_for_row);

			// d_t_3 -> Γ(t) = (T/60) lb(t) e(t)
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "Gamma_%d", t);
			EBR_milp_monomial(varnames_for_row, vals_for_row, 1, -(_this->IBR[IBR_BUY_LOW]), "e_%d", t);
			sprintf(var_name, "d_%d_3", t);
			EBR_milp_row_from_monomials(stream_conv, BOUNDS_TYPE_FIXED, 0, 0, var_name, 0, 2, varnames_for_row, vals_for_row);
		}
	}

	/* OBJECTIVE FUNCTION */
	for (t = 1; t <= TIMESLOTS_AHEAD; t++) {
		sprintf(obj_function, "%s%sGamma_%d", t > 1? obj_function : "", t > 1? " + " : "", t);
	}

	fclose(stream_vars_bnds);
	fclose(stream_vars_list);
	fclose(stream_conv);

	/* Convexing */
	sprintf(filename, "%d.%.0f.%d", myPid, timestamp, timeslot);
	if (EBR_milp_invoke_convexer(filename, MILP_DIR, sql, MAX_BUF_LEN) < 0) {
		fprintf(stderr, "Unrecoverable error in invoking convexer\n");
		exit(1);
	}
	Debug_out(DEBUG, "MILP Convexed\n");

	/* Creating MILP file */
	assert(snprintf(sql, MAX_BUF_LEN, "echo \"Minimize\" > %s/%d.%.0f.%d.lp; echo \"  obj: %s\" >> %s/%d.%.0f.%d.lp; echo \"Subject To\" >> %s/%d.%.0f.%d.lp; cat %s/%d.%.0f.%d.clp.res >> %s/%d.%.0f.%d.lp; echo \"Bounds\" >> %s/%d.%.0f.%d.lp; cat %s/%d.%.0f.%d.lp.bnds >> %s/%d.%.0f.%d.lp; echo \"Binary\" >> %s/%d.%.0f.%d.lp; cat %s/%d.%.0f.%d.lp.bool >> %s/%d.%.0f.%d.lp; echo \"End\" >> %s/%d.%.0f.%d.lp; ", MILP_DIR, myPid, timestamp, timeslot, obj_function, MILP_DIR, myPid, timestamp, timeslot, MILP_DIR, myPid, timestamp, timeslot, MILP_DIR, myPid, timestamp, timeslot, MILP_DIR, myPid, timestamp, timeslot, MILP_DIR, myPid, timestamp, timeslot, MILP_DIR, myPid, timestamp, timeslot, MILP_DIR, myPid, timestamp, timeslot, MILP_DIR, myPid, timestamp, timeslot, MILP_DIR, myPid, timestamp, timeslot, MILP_DIR, myPid, timestamp, timeslot, MILP_DIR, myPid, timestamp, timeslot) < MAX_BUF_LEN);
	system(sql);
	Debug_out(DEBUG, "Created MILP input file '%s/%d.%.0f.%d.lp'\n", MILP_DIR, myPid, timestamp, timeslot);

	/* return filename without dir and without extension */
	return filename;
}


/**
 *	MILP for Substation
 *	Return MILP file name on success, NULL otherwise
 */
char * EBR_create_milp_for_control_subst(EBR *_this, int timeslot, double forecast[FORECASTS_NUMBER][TIMESLOTS_AHEAD], double * measurements)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	Debug_assert(DEBUG, _this->tool == TOOL_SUBST_CTRL, "!(_this->tool == TOOL_SUBST_CTRL)\n");

	static char * varnames_for_row[TIMESLOTS_AHEAD];
	static char obj_function[20*TIMESLOTS_AHEAD];
	static double vals_for_row[TIMESLOTS_AHEAD];

	double d_minus_r;

	char * filename = NULL;
	double maxBattery, minBattery, maxCharge, maxDischarge, maxChSpeed, maxDischSpeed;
	int ts_len, ts_len_tmp;
	int t, i;
	time_t timestamp_time;
	double timestamp;
	int howmany;
	int isThereBattery;

	FILE * stream_milp;

	pid_t myPid = getpid();

	isThereBattery = Battery_getCapacity(_this->battery);

	time(&timestamp_time);
	timestamp = difftime(timestamp_time, 0);

	/* STREAMS */
	for (ts_len = 0, ts_len_tmp = timeslot;
		 ts_len_tmp > 0;
		 ts_len++, ts_len_tmp = ts_len_tmp/10);
	filename = (char *) calloc(strlen(MILP_DIR) + ts_len + 100, sizeof(char));
	sprintf(filename, "%s/%d.%.0f.%d.lp", MILP_DIR, myPid, timestamp, timeslot);
	if (!(stream_milp = fopen(filename, "w"))) {
		fprintf(stderr, "Impossible to open file %s: ", filename);
		perror("");
		exit(1);
	}


	/* OBJECTIVE FUNCTION */
	for (t = 1; t <= TIMESLOTS_AHEAD; t++) {
		sprintf(obj_function, "%s%sGamma_%d", t > 1? obj_function : "", t > 1? " + " : "", t);
	}

	fprintf(stream_milp, "Minimize\n\tobj: %s\n", obj_function);


	/* CONSTRAINTS */

	fprintf(stream_milp, "Subject To\n");

	/* Battery */
	if (isThereBattery) {
		maxCharge = Battery_getChargeFactor(_this->battery);
		maxChSpeed = Battery_getChargeSpeed(_this->battery);
		maxDischSpeed = Battery_getDischargeSpeed(_this->battery);
		/* b_i(2) = b_i(1) + T/60 ( alpha_i * a+_i(1) - a-_i(1) ) */
		/* b_i(2) - T/60 * alpha_i * a+_i(1) + T/60 * a-_i(1) = b_i(1) (= const) */
		EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "b_1_2");
		EBR_milp_monomial(varnames_for_row, vals_for_row, 1, -maxCharge, "a_1_c_1");
		EBR_milp_monomial(varnames_for_row, vals_for_row, 2, 1, "a_1_d_1");
		EBR_milp_row_from_monomials(stream_milp, BOUNDS_TYPE_FIXED, measurements[MEAS_BATTERY], measurements[MEAS_BATTERY], NULL, 0, 3, varnames_for_row, vals_for_row);
		for (t = 1; t <= TIMESLOTS_AHEAD; t++) {
			/* b_i(t+1) = b_i(t) + T/60 ( alpha_i * a+_i(t) - a-_i(t) ) */
			/* b_i(t+1) - b_i(t) - T/60 * alpha_i * a+_i(t) + T/60 * a-_i(t) = 0 */
			if (t > 1) {
				EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "b_1_%d", t+1);
				EBR_milp_monomial(varnames_for_row, vals_for_row, 1, -1, "b_1_%d", t);
				EBR_milp_monomial(varnames_for_row, vals_for_row, 2, -maxCharge, "a_1_c_%d", t);
				EBR_milp_monomial(varnames_for_row, vals_for_row, 3, 1, "a_1_d_%d", t);
				EBR_milp_row_from_monomials(stream_milp, BOUNDS_TYPE_FIXED, 0, 0, NULL, 0, 4, varnames_for_row, vals_for_row);
			}

			/*     y_it -> a+_i(t) = 0 */
			/*
			 *		translates into two constraints:
			 *		1) maxChSpeed y_it + a+_i(t) <= maxChSpeed
			 *		2) -a+_i(t) <= 0
			 */
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, maxChSpeed, "y_1_%d", t);
			EBR_milp_monomial(varnames_for_row, vals_for_row, 1, 1, "a_1_c_%d", t);
			EBR_milp_row_from_monomials(stream_milp, BOUNDS_TYPE_UB, 0, maxChSpeed, NULL, 0, 2, varnames_for_row, vals_for_row);
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, -1, "a_1_c_%d", t);
			EBR_milp_row_from_monomials(stream_milp, BOUNDS_TYPE_UB, 0, 0, NULL, 0, 1, varnames_for_row, vals_for_row);

			/* not y_it -> a-_i(t) = 0 */
			/*
			 *		translates into two constraints:
			 *		a+_i(t) - maxDischSpeed y_it <= 0
			 *		-a+_i(t) <= 0
			 */
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "a_1_d_%d", t);
			EBR_milp_monomial(varnames_for_row, vals_for_row, 1, -maxDischSpeed, "y_1_%d", t);
			EBR_milp_row_from_monomials(stream_milp, BOUNDS_TYPE_UB, 0, 0, NULL, 0, 2, varnames_for_row, vals_for_row);
			EBR_milp_monomial(varnames_for_row, vals_for_row, 0, -1, "a_1_d_%d", t);
			EBR_milp_row_from_monomials(stream_milp, BOUNDS_TYPE_UB, 0, 0, NULL, 0, 1, varnames_for_row, vals_for_row);
		}
	}

	/* Energy balancing */
	/* e(t) + (beta_1 * a-_1(t)) + r(t) = d(t) + sum_{i in 1,2}(a+_i(t)) */
	/* e(t) + (beta_1 * a-_1(t)) - sum_{i in 1,2}(a+_i(t)) = d(t) - r(t) */
	/* maxDischarge = discharge factor of battery */
	maxDischarge = Battery_getDischargeFactor(_this->battery);
	for (t = 1; t <= TIMESLOTS_AHEAD; t++) {
		howmany = 0;
		if (t == 1) {
			d_minus_r = measurements[MEAS_CONSUMPTION];
		}
		else {
			d_minus_r = forecast[FORECASTS_CONSUMPTION][t-2];
		}
		EBR_milp_monomial(varnames_for_row, vals_for_row, howmany++, 1, "e_%d", t);
		if (isThereBattery) {
			EBR_milp_monomial(varnames_for_row, vals_for_row, howmany++, maxDischarge, "a_1_d_%d", t);
			EBR_milp_monomial(varnames_for_row, vals_for_row, howmany++, -1, "a_1_c_%d", t);
		}
		EBR_milp_row_from_monomials(stream_milp, BOUNDS_TYPE_FIXED, d_minus_r, d_minus_r, NULL, 0, howmany, varnames_for_row, vals_for_row);
	}

	/* Mean Energy in Window, forall timeslot >= _this->energyWindowSubst_l */
	if (_this->energyWindowSubst_index == _this->energyWindowSubst_l - 2) {
		double ews_m1 = 1.0/((double)_this->energyWindowSubst_l);
		Debug_out(DEBUG, "1/delta = %lg\n", ews_m1);
		for (t = 1; t <= TIMESLOTS_AHEAD; t++) {
			/*
			 *	let \delta = _this->energyWindowSubst_l
			 *	let P_S = _this->maxEnergy
			 *	{1 \over \delta} \sum_{t - \delta < i <= t}{e(i)} <= P_S
			 */
			/* compututation of mean of stored numbers for energy */
			double rhs = 0.0;
			for (i = t - 1; i <= _this->energyWindowSubst_index; i++) {
				rhs += (ews_m1 * _this->energyWindowSubst[i]);
			}
			Debug_out(DEBUG, "Mean of Energy Window (%d .. %d) = %lg\n", t - 1, _this->energyWindowSubst_index, rhs);
			/* denotation of mean of remaining variables */
			howmany = 0;
			EBR_milp_monomial(varnames_for_row, vals_for_row, howmany++, ews_m1, "e_1");
			for (i = 2; i <= t; i++) {
				EBR_milp_monomial(varnames_for_row, vals_for_row, howmany++, ews_m1, "e_%d", i);
			}
			rhs = _this->maxEnergy - rhs;
			EBR_milp_row_from_monomials(stream_milp, BOUNDS_TYPE_UB, rhs, rhs, NULL, 0, howmany, varnames_for_row, vals_for_row);
			Debug_out(DEBUG, "Added constraint on Energy Window for t = %d\n", t);
		}
	}

	/* Cost */
	for (t = 1; t <= TIMESLOTS_AHEAD; t++) {
		// Γ(t) = (T/60) c(t) e(t)
		EBR_milp_monomial(varnames_for_row, vals_for_row, 0, 1, "Gamma_%d", t);
		EBR_milp_monomial(varnames_for_row, vals_for_row, 1, -(_this->SUBSToU[t-1]), "e_%d", t);
		EBR_milp_row_from_monomials(stream_milp, BOUNDS_TYPE_FIXED, 0, 0, NULL, 0, 2, varnames_for_row, vals_for_row);
	}


	/* REAL VALUED VARIABLES */

	fprintf(stream_milp, "Bounds\n");

	/**
	 *	For all energy storage appliances i, for all timeslots in the future
	 *	- b_i(t) in [20% B_iu, B_iu]
	 *		- b_i(1) fixed as measured
	 *	- a+_i(t) in [0, C_iu]
	 *	- a-_i(t) in [0, C_il]
	 *	- y_it	in B
	 */
	/* Battery = 1 */
	maxBattery = Battery_getCapacity(_this->battery);
	if (isThereBattery) {
		minBattery = 20.0*maxBattery/100.0;
		maxCharge = Battery_getChargeSpeed(_this->battery);
		maxDischarge = Battery_getDischargeSpeed(_this->battery);
		for (t = 1; t <= TIMESLOTS_AHEAD; t++) {
			if (t > 1) {
				/* b_i(t) */
				fprintf(stream_milp, "\t%lg <= b_1_%d <= %lg\n", minBattery, t, maxBattery);
			}
			/* a+_i(t) charge */
			fprintf(stream_milp, "\t0 <= a_1_c_%d <= %lg\n", t, maxCharge);
			/* a-_i(t) discharge */
			fprintf(stream_milp, "\t0 <= a_1_d_%d <= %lg\n", t, maxDischarge);
			/* y_it */
			fprintf(stream_milp, "\t0 <= y_1_%d <= 1\n", t);
		}
		/* Add also last TIMESLOTS_AHEAD+1 battery state of charge variable */
		fprintf(stream_milp, "\t%lg <= b_1_%d <= %lg\n", minBattery, TIMESLOTS_AHEAD+1, maxBattery);
	}

	/**
	 *	Energy from (to) grid
	 *	- e(t) >= min (usually min = 0, but taken from input file)
	 */
	for (t = 1; t <= TIMESLOTS_AHEAD; t++) {
		fprintf(stream_milp, "\t%lg <= e_%d\n", _this->minEnergy, t);
	}

	/**
	 *	Cost function
	 *	- Gamma(t)	free
	 */
	for (t = 1; t <= TIMESLOTS_AHEAD; t++) {
		/* Gamma_t */
		fprintf(stream_milp, "\tGamma_%d free\n", t);
	}

	/* BINARY VARIABLES */
	fprintf(stream_milp, "Binary\n");

	if (isThereBattery) {
		for (t = 1; t <= TIMESLOTS_AHEAD; t++) {
			/* y_it */
			fprintf(stream_milp, "\ty_1_%d\n", t);
		}
	}

	fprintf(stream_milp, "End\n");

	fflush(stream_milp);
	fclose(stream_milp);

	sprintf(filename, "%d.%.0f.%d", myPid, timestamp, timeslot);

	Debug_out(DEBUG, "Created MILP input file '%s'\n", filename);
	
	/* return filename without dir and without extension */
	return filename;
}



#if 0
int EBR_decideCommands_greedy(EBR *_this, int timeslot, double forecast[FORECASTS_NUMBER][TIMESLOTS_AHEAD], double *measurements, double *commands, double *gridEnergy)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");

	double chargeSpeed, dischargeSpeed, factor;
	double capacity;
	double soc;
	double needed, maxAvail;
	/* only need next timeslot for this greedy strategy */
	double consMinusProd = forecast[FORECASTS_CONSUMPTION][0] - forecast[FORECASTS_PRODUCTION][0];

	Debug_out(DEBUG, "forecast[%d][0] = %lg (consumption)\n", FORECASTS_CONSUMPTION, forecast[FORECASTS_CONSUMPTION][0]);
	Debug_out(DEBUG, "forecast[%d][0] = %lg (production)\n", FORECASTS_PRODUCTION, forecast[FORECASTS_PRODUCTION][0]);
	Debug_out(DEBUG, "forecast[%d][0] = %lg (PHEV)\n", FORECASTS_PHEV, forecast[FORECASTS_PHEV][0]);
	Debug_out(DEBUG, "consMinusProd = %lg\n", consMinusProd);

	commands[CMDS_BATTERY] = 0;
	commands[CMDS_PHEV] = 0;
	*gridEnergy = consMinusProd;

	/** Greedy strategy */

	/** if possible, charge batteries as much as possible */
	if (_this->pricePolicy[PRICEPOLICY_MAX][0] > consMinusProd) {
		/* if present in next timeslot, then charge PHEV as much as possible */
		if (forecast[FORECASTS_PHEV][0] > 0) {
			soc = measurements[MEAS_PHEV];
			capacity = Battery_getCapacity(_this->PHEV);
			chargeSpeed = Battery_getChargeSpeed(_this->PHEV);
			factor = Battery_getChargeFactor(_this->PHEV);

			Debug_out(DEBUG, "1. soc = %lg\n", soc);

			if (soc < capacity) {
				needed = (capacity - soc)/factor < chargeSpeed? (capacity - soc)/factor : chargeSpeed;
				maxAvail = (_this->pricePolicy[PRICEPOLICY_MAX][0] - consMinusProd)/factor;

				Debug_out(DEBUG, "1. needed = %lg\n", needed);
				Debug_out(DEBUG, "1. maxAvail = %lg\n", maxAvail);

				commands[CMDS_PHEV] = maxAvail <= needed? maxAvail : needed;
				*gridEnergy += commands[CMDS_PHEV];
			}
		}
		/* then, charge battery as much as possible */
		if (*gridEnergy < _this->pricePolicy[PRICEPOLICY_MAX][0]) {
			soc = measurements[MEAS_BATTERY];
			capacity = Battery_getCapacity(_this->battery);
			chargeSpeed = Battery_getChargeSpeed(_this->battery);
			factor = Battery_getChargeFactor(_this->battery);

			Debug_out(DEBUG, "2. soc = %lg\n", soc);

			if (soc < capacity) {
				needed = (capacity - soc)/factor < chargeSpeed? (capacity - soc)/factor : chargeSpeed;
				maxAvail = (_this->pricePolicy[PRICEPOLICY_MAX][0] - consMinusProd)/factor;

				Debug_out(DEBUG, "2. needed = %lg\n", needed);
				Debug_out(DEBUG, "2. maxAvail = %lg\n", maxAvail);

				commands[CMDS_BATTERY] = maxAvail <= needed? maxAvail : needed;
				*gridEnergy += commands[CMDS_BATTERY];
			}
		}

	}
	else {
		/* use required or available battery */
		soc = measurements[MEAS_BATTERY];
		dischargeSpeed = Battery_getDischargeSpeed(_this->battery);
		factor = Battery_getDischargeFactor(_this->battery);

		Debug_out(DEBUG, "3. soc = %lg\n", soc);

		needed = (consMinusProd - _this->pricePolicy[PRICEPOLICY_MAX][0])/factor;
		maxAvail = dischargeSpeed/factor < soc? dischargeSpeed/factor : soc;

		Debug_out(DEBUG, "3. needed = %lg\n", needed);
		Debug_out(DEBUG, "3. maxAvail = %lg\n", maxAvail);

		commands[CMDS_BATTERY] = maxAvail <= needed? -maxAvail : -needed;
		*gridEnergy -= commands[CMDS_BATTERY];

	}

	Debug_out(DEBUG, "*gridEnergy = %lg\n", *gridEnergy);
	Debug_out(DEBUG, "commands[%d] = %lg\n", CMDS_BATTERY, commands[CMDS_BATTERY]);
	Debug_out(DEBUG, "commands[%d] = %lg\n", CMDS_PHEV, commands[CMDS_PHEV]);
	
	return 0;
}
#endif

