#define _BSD_SOURCE

#include <EUMF-K.h>
#include <Debug.h>
#include <StringUtils.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>


#define DEBUG "EBR"




static void _EBR_forecast_usePast(EBR *_this, int timeslot, double *consumption, double *production, double *PHEV);

static int _EBR_forecast_useHistory(EBR *_this, double *consumption, double *production, double *PHEV);




int EBR_forecast(EBR *_this, int timeslot, double *consumption, double *production, double *PHEV)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	Debug_assert(DEBUG, consumption != NULL, "consumption == NULL\n");
	Debug_assert(DEBUG, production != NULL, "production == NULL\n");
	Debug_assert(DEBUG, 1 <= timeslot && timeslot <= TIMESLOTS_IN_A_DAY, "timeslot out of bounds\n");
	Debug_assert(DEBUG, _this->tool == TOOL_CTRL || _this->tool == TOOL_EVAL || _this->tool == TOOL_SUBST_CTRL, "!(_this->tool == TOOL_CTRL || _this->tool == TOOL_EVAL || _this->tool == TOOL_SUBST_CTRL)\n");

	int res = 0;

	if (_this->tool == TOOL_EVAL) {
		res = _EBR_forecast_useHistory(_this, consumption, production, PHEV);
	}
	else { /* TOOL*_CTRL */
		_EBR_forecast_usePast(_this, timeslot, consumption, production, PHEV);
	}

	return res;
}



static void _EBR_forecast_usePast(EBR *_this, int timeslot, double *consumption, double *production, double *PHEV)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	Debug_assert(DEBUG, consumption != NULL, "consumption == NULL\n");
	Debug_assert(DEBUG, production != NULL, "production == NULL\n");
	Debug_assert(DEBUG, 1 <= timeslot && timeslot <= TIMESLOTS_IN_A_DAY, "timeslot out of bounds\n");
	Debug_assert(DEBUG, _this->tool == TOOL_CTRL || _this->tool == TOOL_SUBST_CTRL, "!(_this->tool == TOOL_CTRL || _this->tool == TOOL_SUBST_CTRL)\n");

	int i, index, j, absence;
	double tolerableAbsence;
	int pastLength = _this->pastLength[timeslot-1];

	Debug_out(DEBUG, "pastLength = %d\n", pastLength);

	
	/* i writes forcast modulo TIMESLOTS_AHEAD ([0] is forecast for next timeslot...) */
	for (i = 0; i < TIMESLOTS_AHEAD; i++) {
		/* index reads in _this->past[] modulo TIMESLOTS_IN_A_DAY */
		index = (timeslot-1) + i + 1;
		if (index >= TIMESLOTS_IN_A_DAY) {
			/* index -= TIMESLOTS_IN_A_DAY; */
			index = index%TIMESLOTS_IN_A_DAY;
		}
		consumption[i] = 0.0;
		if (_this->tool == TOOL_CTRL) {
			production[i] = 0.0;
		}
		if (pastLength > 0) {
//			Debug_out(DEBUG, "** PAST RECORDED FOR TIMESLOT %d **", index+1);
//			Debug_perform(DEBUG, EBR_pastDump(_this, index+1));
			for (j = 0; j < pastLength; j++) {
				consumption[i] += _this->pastDistribution[j] * _this->past[FORECASTS_CONSUMPTION][index][j];
				if (_this->tool == TOOL_CTRL) {
					production[i] += _this->pastDistribution[j] * _this->past[FORECASTS_PRODUCTION][index][j];
				}
			}
		}
	}

	if (_this->tool == TOOL_CTRL && _this->PHEV) {
		tolerableAbsence = TOLERABLE_PHEV_ABSENCE * (double)pastLength;
		//	Debug_out(DEBUG, "tolerableAbsence = %lg\n", tolerableAbsence);

		/* i writes forcast modulo TIMESLOTS_AHEAD ([0] is forecast for next timeslot...) */
		for (i = 0; i < TIMESLOTS_AHEAD; i++) {
			/* index reads in _this->past[] modulo TIMESLOTS_IN_A_DAY */
			index = (timeslot-1) + i + 1;
			if (index >= TIMESLOTS_IN_A_DAY) {
				/* index -= TIMESLOTS_IN_A_DAY; */
				index = index%TIMESLOTS_IN_A_DAY;
			}
			//		Debug_out(DEBUG, "index = %d\n", index);
			PHEV[i] = 0.0;
			if (pastLength > 0) {
				absence = 0;
				for (j = 0; j < pastLength && absence <= tolerableAbsence; j++) {
					/* _this->pastPHEV[index][j] is -1 if PHEV is not present */
					if (_this->past[FORECASTS_PHEV][index][j] >= 0) {
						PHEV[i] += _this->past[FORECASTS_PHEV][index][j];
					}
					else {
						absence++;
					}
				}
				//			Debug_out(DEBUG, "absence = %d\n", absence);
				PHEV[i] = absence <= tolerableAbsence? PHEV[i]/(double)(pastLength - absence) : -1;
			}
			else
				PHEV[i] = -1;
			//		Debug_out(DEBUG, "PHEV[%d] = %lg\n", i, PHEV[i]);
		}
	}
}



static int _EBR_forecast_useHistory(EBR *_this, double *consumption, double *production, double *PHEV)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	Debug_assert(DEBUG, consumption != NULL, "consumption == NULL\n");
	Debug_assert(DEBUG, production != NULL, "production == NULL\n");
	Debug_assert(DEBUG, PHEV != NULL, "PHEV == NULL\n");
	Debug_assert(DEBUG, _this->tool == TOOL_EVAL, "!(_this->tool == TOOL_EVAL)\n");

	static char line[MAX_BUF_LEN];
	char *lineTmp;
	char * token = NULL;
	int t;
	int h = TIMESLOTS_AHEAD - 1;

	if (_this->useHistory < TIMESLOTS_AHEAD) {
		/* first time we must discard first measure because we have to forecast starting from next time step */
		h = -1;
	}

	do {

		for (t = 0; t < TIMESLOTS_AHEAD-1; t++) {
			consumption[t] = consumption[t+1];
			production[t]  = production[t+1];
			PHEV[t] = PHEV[t+1];
		}

		/* read next line from file and put it as last forecast */
		if (!fgets(line, MAX_BUF_LEN, _this->useHistoryFile_fd)) {
			Debug_out(DEBUG, "No more lines to read from history\n");
			return -1;
		}
		if (line[strlen(line)-1] == '\n') {
			line[strlen(line)-1] = '\0';
		}
		Debug_out(DEBUG, "Retrieved profiles : %s\n", line);

		/* starting */
		lineTmp = StringUtils_clone(line);
		token = strsep(&lineTmp, ",");
		if (!token) {
			perror("Problems tokenizing file line");
			exit(1);
		}
		/* ending */
		token = strsep(&lineTmp, ",");
		if (!token) {
			perror("Problems tokenizing file line");
			exit(1);
		}
		/* consumption */
		token = strsep(&lineTmp, ",");
		if (!token) {
			perror("Problems tokenizing file line");
			exit(1);
		}
		if (StringUtils_toDouble(token, &consumption[TIMESLOTS_AHEAD-1]) != 0) {
			Debug_out(DEBUG, "Problems reading consumption, value '%s' not accepted\n", token);
			exit(1);
		}
		/* production */
		token = strsep(&lineTmp, ",");
		if (!token) {
			perror("Problems tokenizing file line");
			exit(1);
		}
		if (StringUtils_toDouble(token, &production[TIMESLOTS_AHEAD-1]) != 0) {
			Debug_out(DEBUG, "Problems reading production, value '%s' not accepted\n", token);
			exit(1);
		}
		/* check phev presence */
		token = strsep(&lineTmp, ",");
		if (!token) {
			perror("Problems tokenizing file line");
			exit(1);
		}
		if (!StringUtils_equals(token, "")) {
			if (StringUtils_toDouble(token, &PHEV[TIMESLOTS_AHEAD-1]) != 0) {
				Debug_out(DEBUG, "Problems reading phev, value %s not accepted\n", token);
				exit(1);
			}
		}
		else {
			Debug_out(DEBUG, "PHEV not present\n");
			PHEV[TIMESLOTS_AHEAD-1] = -1;
		}

		/* HERE WE DO NOT CARE ABOUT PHEV_READY_HOURS OR PHEV_CHARGE BECAUSE WE READ FROM HOUSE */

		/* increase useHistory, if need be */
		if (_this->useHistory < TIMESLOTS_AHEAD) {
			(_this->useHistory)++;
		}

//		free(lineTmp);

		h++;

	} while (h < TIMESLOTS_AHEAD);

	return 0;
}


void EBR_readPHEVCommand(EBR *_this, double * commands)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	Debug_assert(DEBUG, commands != NULL, "commands == NULL\n");
	Debug_assert(DEBUG, _this->tool == TOOL_HISTORY ||
				 (_this->tool == TOOL_CTRL && EBR_isPastTraining(_this)), "!(_this->tool == TOOL_HISTORY  || (_this->tool == TOOL_CTRL && EBR_isPastTraining(_this)))\n");

	static char line[MAX_BUF_LEN];
	char *lineTmp;
	char * token = NULL;

	/* read next line from file and put it as last forecast */
	if (!fgets(line, MAX_BUF_LEN, _this->useHistoryFile_fd)) {
		Debug_out(DEBUG, "No more lines to read from history\n");
		return;
	}
	if (line[strlen(line)-1] == '\n') {
		line[strlen(line)-1] = '\0';
	}
	Debug_out(DEBUG, "Retrieved profiles : %s\n", line);

	/* starting */
	lineTmp = StringUtils_clone(line);
	token = strsep(&lineTmp, ",");
	if (!token) {
		perror("Problems tokenizing file line");
		exit(1);
	}
	/* ending */
	token = strsep(&lineTmp, ",");
	if (!token) {
		perror("Problems tokenizing file line");
		exit(1);
	}
	/* consumption */
	token = strsep(&lineTmp, ",");
	if (!token) {
		perror("Problems tokenizing file line");
		exit(1);
	}
	/* production */
	token = strsep(&lineTmp, ",");
	if (!token) {
		perror("Problems tokenizing file line");
		exit(1);
	}
	/* check phev presence */
	token = strsep(&lineTmp, ",");
	if (!token) {
		perror("Problems tokenizing file line");
		exit(1);
	}
	/* phev charging kwh */
	token = strsep(&lineTmp, ",");
	if (!token) {
		perror("Problems tokenizing file line");
		exit(1);
	}
	if (!StringUtils_equals(token, "")) {
		if (StringUtils_toDouble(token, &commands[CMDS_PHEV]) != 0) {
			Debug_out(DEBUG, "Problems reading phev charging, value %s not accepted\n", token);
			exit(1);
		}
	}
	else {
		commands[CMDS_PHEV] = 0;
	}

	/* HERE WE DO NOT CARE ABOUT PHEV_READY_HOURS BECAUSE WE ARE IN SIMULATION */

//	free(lineTmp);
}



void EBR_substEnergyWindowPush(EBR * _this, double energy)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	Debug_assert(DEBUG, _this->tool == TOOL_SUBST_CTRL, "tool != TOOL_SUBST_CTRL\n");

	/* shift array to the left, if need be */
	if (_this->energyWindowSubst_index == _this->energyWindowSubst_l - 2) {
		int i;
		for (i = 1; i <= _this->energyWindowSubst_index; i++) {
			_this->energyWindowSubst[i-1] = _this->energyWindowSubst[i];
		}
		Debug_out(DEBUG, "Shifted Energy Window one step to the left\n");
	}

	/* increase last index, until it reaches the end */
	if (_this->energyWindowSubst_index < _this->energyWindowSubst_l - 2) {
		(_this->energyWindowSubst_index)++;
	}

	/* store energy in last position */
	_this->energyWindowSubst[_this->energyWindowSubst_index] = energy;
	Debug_out(DEBUG, "Assigned index %d of Energy Window to %lg\n", _this->energyWindowSubst_index, energy);
}

