#include <EUMF-K.h>
#include <Debug.h>
#include <CommandLineInterface.h>
#include <HECH.h>


#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <unistd.h>
#include <time.h>


#define DEBUG "EBR"



int main(int argc, char **argv)
{
	CommandLineInterface * cliPtr = NULL;

	HECH * myHECH = NULL;
	EBR *this = NULL;

	int timeslot = 0;
	double energyCost = 0;
	double timeslotEnergyCost = 0;
	long timeslots_num = 0;

	time_t start, end;

	/* measures from House (enum HECH.h:Measures) */
	double measurements[MEAS_NUMBER] = { 0 };

	/* charge/discharge/nop to battery and phev (enum HECH.h:Commands)  */
	double commands[CMDS_NUMBER] = { 0 };

	/* forecast of energy exchange with grid */
	double energyForecastedForThisTimeslot = 0;

	/**
	 *	Forecast for next TIMESLOTS_AHEAD timeslots
	 *	for enum EBR.h:Forecasts
	 */
	double forecast[FORECASTS_NUMBER][TIMESLOTS_AHEAD] = {{0}};


	/* Command line interface */
	cliPtr = EBR_CLI(argc, argv);

	/* Create this EBR object */
	this = EBR_new(cliPtr);

	/* Set-up connect to the HECH */
	myHECH = EBR_houseConnect(this->configHECH);
	if (!myHECH) {
		Debug_out(DEBUG, "Failed to connect to the House server\n");
		exit(1);
	}
	Debug_perform(DEBUG, HECH_dumpInfo(myHECH, stdout));


	/* Timer_start("EBR main algorithm"); */
	time(&start);

	EBR_printBanner(stdout);

	timeslot = 1;

	while (1) {

		int answer, i;

		for (i = 0; i < CMDS_NUMBER; i++) {
			commands[i] = 0;
		}

		Debug_out(DEBUG, "Timeslot = %d\n", timeslot);


		/* Get current hour, for ToU purposes */
		if (this->tariff == TARIFF_TOU) {
			EBR_updateCurrentHour(this);
			Debug_out(DEBUG, "Current hour = %d\n", this->currentHour);
		}


		/* Read measurements from the House */
		Debug_out(DEBUG, "Read measurements from House server\n");
		answer = HECH_requestMeasurements(myHECH, measurements);
		if (answer == HECH_getControlStopSignal(myHECH)) {
			Debug_out(DEBUG, "Control STOP signal sent from House server\n");
			break;
		}
		else if (answer == -2) {
			Debug_out(DEBUG, "Cannot get measurements from House server\n");
			break;
		}
		Debug_out(DEBUG, "Measurements from House : energy from grid = %lg, consumption = %lg, production = %lg, battery SoC = %lg, PHEV SoC = %lg, PHEV ready hours = %lg\n", measurements[MEAS_ENERGY], measurements[MEAS_CONSUMPTION], measurements[MEAS_PRODUCTION], measurements[MEAS_BATTERY], measurements[MEAS_PHEV], measurements[MEAS_PHEV_READY_HOURS]);


		/* Reads current price policy, if tariff is IBR  */
		if (this->tariff == TARIFF_IBR) {
			Debug_out(DEBUG, "Reads current price policy\n");
			answer = EBR_stepPricePolicy(this);
			if (answer < 0) {
				Debug_out(DEBUG, "Cannot get current price policy\n");
				break;
			}
			Debug_out(DEBUG, "Current price policy:\n");
			Debug_perform(DEBUG, EBR_dumpPricePolicy(this));
		}
		else if (this->tariff == TARIFF_SUBST) {
			Debug_out(DEBUG, "Reads current Substation ToU tariff\n");
			answer = EBR_stepSubstTariff(this);
			if (answer < 0) {
				Debug_out(DEBUG, "Cannot get current Substation ToU tariff\n");
				break;
			}
			Debug_out(DEBUG, "Current Substation ToU tariff:\n");
			Debug_perform(DEBUG, EBR_dumpSubstTariff(this));
		}


		if (this->tool == TOOL_CTRL || this->tool == TOOL_SUBST_CTRL) {
			/* Update PAST for consumption, production, and PHEV for forecasting  */
			EBR_pastPush(this, timeslot, measurements[MEAS_CONSUMPTION], measurements[MEAS_PRODUCTION], measurements[MEAS_PHEV]);
			Debug_out(DEBUG, "Measures added to PAST for learning on future\n");
			Debug_out(DEBUG, "** PAST RECORDED FOR CURRENT TIMESLOT **\n");
			Debug_perform(DEBUG, EBR_pastDump(this, timeslot));
		}


		/* Get real PHEV command if in simulation or if training TOOL_CTRL */
		if ( this->tool == TOOL_HISTORY ||
			(this->tool == TOOL_CTRL && EBR_isPastTraining(this)) ) {
			double energy;
			EBR_readPHEVCommand(this, commands);
			Debug_out(DEBUG, "Read command PHEV = %lg\n", commands[CMDS_PHEV]);
			energy = EBR_energyFromGrid(this, measurements, commands);
			EBR_energyPrice(this, energy, &timeslotEnergyCost);
		}

		/* Get forecast and decide, if training is finished */
		if ( this->tool == TOOL_EVAL ||
			(this->tool == TOOL_CTRL && !EBR_isPastTraining(this)) ||
			(this->tool == TOOL_SUBST_CTRL && !EBR_isPastTraining(this)) ) {
			answer = EBR_forecast(this, timeslot, forecast[FORECASTS_CONSUMPTION], forecast[FORECASTS_PRODUCTION], forecast[FORECASTS_PHEV]);
			if (answer < 0) {
				Debug_out(DEBUG, "No more future to forecast. Experiment ends here.\n");
				break;
			}
			Debug_out(DEBUG, "Forecast at timeslot %d\n", timeslot);
			Debug_perform(DEBUG, EBR_printForecasts(timeslot, forecast));

			/* Decide, when needed */
			Debug_out(DEBUG, "Decide commands for battery and PHEV, if any\n");
			/* answer = EBR_decideCommands(this, timeslot, forecast, measurements, commands, &energyForecastedForThisTimeslot, &timeslotEnergyCost); */
			answer = -1;
			if (answer < 0) {
				double energy;
				Debug_out(DEBUG, "Cannot decide, setting commands and energy to 0\n");
				energyForecastedForThisTimeslot = 0;
				commands[CMDS_BATTERY] = 0;
				commands[CMDS_PHEV] = 0;
				energy = EBR_energyFromGrid(this, measurements, commands);
				EBR_energyPrice(this, energy, &timeslotEnergyCost);
			}
			Debug_out(DEBUG, "Decided commands : battery = %lg, PHEV = %lg\n", commands[CMDS_BATTERY], commands[CMDS_PHEV]);
		}


		/* For Substation only, store energy in Window */
		if (this->tool == TOOL_SUBST_CTRL) {
			if (EBR_isPastTraining(this)) {
				commands[CMDS_BATTERY] = 0;
				commands[CMDS_PHEV] = 0;
				energyForecastedForThisTimeslot = EBR_energyFromGrid(this, measurements, commands);
				EBR_energyPrice(this, energyForecastedForThisTimeslot, &timeslotEnergyCost);
			}
			EBR_substEnergyWindowPush(this, energyForecastedForThisTimeslot);
			Debug_out(DEBUG, "Energy Window in the past:\n");
			Debug_perform(DEBUG, EBR_dumpEnergyWindow(this));
		}

		/* For Substation simulation */
		if (this->tool == TOOL_SUBST_HISTORY) {
			commands[CMDS_BATTERY] = 0;
			commands[CMDS_PHEV] = 0;
			energyForecastedForThisTimeslot = EBR_energyFromGrid(this, measurements, commands);
			EBR_energyPrice(this, energyForecastedForThisTimeslot, &timeslotEnergyCost);
		}


		/* Send commands to the House */
		Debug_out(DEBUG, "Send commands to House server\n");
		answer = HECH_sendCommands(myHECH, commands);
		if (answer < 0) {
			Debug_out(DEBUG, "Commands not accepted\n");
		}
		Debug_out(DEBUG, "Sent commands to House : battery = %lg, PHEV = %lg\n", commands[CMDS_BATTERY], commands[CMDS_PHEV]);


		/* Log actual measurements with previous commands */
		/* e(t + 1) = d(t + 1) + beta_1 * a_1(t) + beta_2 * a_2(t) + r(t + 1) */
		EBR_logMeasurements(this, measurements, commands, forecast, energyForecastedForThisTimeslot, timeslotEnergyCost);
		Debug_out(DEBUG, "Logged measurements in %s\n", this->loggingFile);


		/* Store total cost */
		energyCost += timeslotEnergyCost;
		timeslots_num++;


		/* Sleep for the requested time */
//		Debug_out(DEBUG, "Sleep for one hour\n");
//		sleep(60*60);

		/* needed for forecast */
		timeslot++;
		if (timeslot > TIMESLOTS_IN_A_DAY) {
			timeslot = 1;
		}
		EBR_nextPastTraining(this);
	}

	/* Timer_end("EBR main algorithm"); */
	time(&end);
	printf("\nExecution time: %lg seconds\n", difftime(end, start));

	printf("\n** TOTAL   ENERGY BILL      = %lg\n", energyCost);
	printf("** MONTHLY ENERGY BILL MEAN = %lg\n", energyCost/((double)timeslots_num/30.0));
	printf("** DAILY   ENERGY BILL MEAN = %lg\n\n", energyCost/(double)timeslots_num);

	if (myHECH != NULL) EBR_houseDisconnect(&myHECH);
	if (this != NULL) EBR_free(&this);
	/* if (cliPtr != NULL) CommandLineInterface_free(&cliPtr); */

	return 0;
}
