#define _GNU_SOURCE

#include <EUMF-K.h>
#include <Debug.h>
#include <StringUtils.h>
#include <CommandLineInterface.h>
#include <Properties.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <math.h>


#define DEBUG "EBR"


static void _EBR_readEnergyFile(EBR *_this, char * energyFile);
static void _EBR_pastDistribution(EBR *_this, int maxPastLength, char *pastDistribution);


EBR * EBR_new(CommandLineInterface *cliPtr)
{
	static char line[MAX_BUF_LEN];
	static char tariffsToUFile[PATH_MAX];
	static char energyFile[PATH_MAX];
	char * pastDistribution = NULL;
	const char * tool;
	const char * tariff;

	EBR * new = (EBR *) calloc(1, sizeof(EBR));
	int i, j;

	/* Get CLI options */
	new->configFile = CommandLineInterface_getStringOption(cliPtr, "configHECH");

	/* Reads HECH parameters from configuration file */
	new->configHECH = Properties_newFromFile(new->configFile, 1024, 1024);
	if (!new->configHECH) {
		perror("Problems reading House configuration file for the HECH");
		exit(1);
	}
	Debug_out(DEBUG, "Configuration file %s:\n", new->configFile);
	Debug_perform(DEBUG, Properties_fprint(new->configHECH, stdout));

	/* Read tool */
	tool = CommandLineInterface_getStringOption(cliPtr, "tool");
	if (StringUtils_equalsIgnoreCase(tool, "EBR-CTRL")) {
		new->tool = TOOL_CTRL;
	}
	else if (StringUtils_equalsIgnoreCase(tool, "EBR-EVAL")) {
		new->tool = TOOL_EVAL;
	}
	else if (StringUtils_equalsIgnoreCase(tool, "EBR-HISTORY")) {
		new->tool = TOOL_HISTORY;
	}
	else if (StringUtils_equalsIgnoreCase(tool, "DAPP-S-CTRL")) {
		new->tool = TOOL_SUBST_CTRL;
	}
	else if (StringUtils_equalsIgnoreCase(tool, "DAPP-S-HISTORY")) {
		new->tool = TOOL_SUBST_HISTORY;
	}
	else {
		fprintf(stderr, "Tool '%s' not supported. Choose one among {EBR-CTRL, EBR-EVAL, EBR-HISTORY, DAPP-S-CTRL, DAPP-S-HISTORY}", tool);
		fflush(NULL);
		exit(1);
	}

	/* Read tariff */
	tariff = CommandLineInterface_getStringOption(cliPtr, "tariff");
	if (StringUtils_equalsIgnoreCase(tariff, "FLAT")) {
		new->tariff = TARIFF_FLAT;
	}
	else if (StringUtils_equalsIgnoreCase(tariff, "IBR")) {
		new->tariff = TARIFF_IBR;
	}
	else if (StringUtils_equalsIgnoreCase(tariff, "ToU")) {
		new->tariff = TARIFF_TOU;
	}
	else if (StringUtils_equalsIgnoreCase(tariff, "SUBST")) {
		new->tariff = TARIFF_SUBST;
	}
	else {
		fprintf(stderr, "Tariff '%s' not supported. Choose one among {FLAT, IBR, ToU, SUBST}", tariff);
		fflush(NULL);
		exit(1);
	}

	/* sanity check on substation tool */
	if (new->tool == TOOL_SUBST_CTRL && new->tariff != TARIFF_SUBST) {
		perror("Tool DAPP-S-CTRL can be only used with tariff SUBST");
		exit(1);
	}
	if (new->tool == TOOL_SUBST_HISTORY && new->tariff != TARIFF_SUBST) {
		perror("Tool DAPP-S-HISTORY can be only used with tariff SUBST");
		exit(1);
	}

	/* retrieve needed arguments */
	if (new->tariff == TARIFF_IBR) {
		realpath(CommandLineInterface_getStringOption(cliPtr, "pricePolicyFile"), new->pricePolicyFile);
		new->IBR[IBR_BUY_LOW] = CommandLineInterface_getDoubleOption(cliPtr, "IBRBuyLow");
		new->IBR[IBR_BUY_HIGH] = CommandLineInterface_getDoubleOption(cliPtr, "IBRBuyHigh");
		new->IBR[IBR_SELL] = CommandLineInterface_getDoubleOption(cliPtr, "IBRSell");
	}
	else if (new->tariff == TARIFF_TOU) {
		realpath(CommandLineInterface_getStringOption(cliPtr, "tariffsToUFile"), tariffsToUFile);
	}
	else if (new->tariff == TARIFF_FLAT) {
		new->FLAT[TOU_BUY] = CommandLineInterface_getDoubleOption(cliPtr, "FLATBuy");
		new->FLAT[TOU_SELL] = CommandLineInterface_getDoubleOption(cliPtr, "FLATSell");

	}
	else if (new->tariff == TARIFF_SUBST) {
		realpath(CommandLineInterface_getStringOption(cliPtr, "tariffsSubstFile"), new->tariffsSubstFile);
	}

	if (new->tool == TOOL_EVAL || new->tool == TOOL_HISTORY || new->tool == TOOL_CTRL || new->tariff == TARIFF_TOU) {
		new->useHistory = 1;
		realpath(CommandLineInterface_getStringOption(cliPtr, "useHistoryFile"), new->useHistoryFile);
	}

	/* Logging and battery/phev configuration parameters */
	realpath(CommandLineInterface_getStringOption(cliPtr, "loggingFile"), new->loggingFile);
	realpath(CommandLineInterface_getStringOption(cliPtr, "batteryFile"), new->batteryFile);
	const char * phev = CommandLineInterface_getStringOption(cliPtr, "phevFile");
	if (phev) {
		realpath(phev, new->phevFile);
	}
	else {
		bzero(new->phevFile, PATH_MAX);
	}


	/* Past */
	if (new->tool == TOOL_CTRL || new->tool == TOOL_SUBST_CTRL) {
		new->maxPastLength = CommandLineInterface_getIntOption(cliPtr, "pastLength");
		pastDistribution = strdup(CommandLineInterface_getStringOption(cliPtr, "pastDistribution"));
		_EBR_pastDistribution(new, new->maxPastLength, pastDistribution);
	}

	/* Energy Window */
	if (new->tool == TOOL_SUBST_CTRL) {
		new->energyWindowSubst_l = CommandLineInterface_getIntOption(cliPtr, "energyWindowSubst");
		/* Window has length equal to (new->energyWindowSubst_l - 1) */
		new->energyWindowSubst = (double *) calloc(new->energyWindowSubst_l - 1, sizeof(double));
		new->energyWindowSubst_index = -1;
	}

	/* Read contractual constraints */
	realpath(CommandLineInterface_getStringOption(cliPtr, "energyFile"), energyFile);
	_EBR_readEnergyFile(new, energyFile);

	Debug_out(DEBUG, "Command Line Interface:\n");
	Debug_perform(DEBUG, CommandLineInterface_fprint_args(cliPtr, stdout));
	Debug_out(DEBUG, "** USING TOOL   : %s\n", tool);
	Debug_out(DEBUG, "** USING TARIFF : %s\n", tariff);

	if (new->tool == TOOL_EVAL || new->tool == TOOL_HISTORY || new->tariff == TARIFF_TOU) {
		Debug_out(DEBUG, "Absolute path of useHistoryFile: %s\n", new->useHistoryFile);
	}
	else if (new->tool == TOOL_CTRL || new->tool == TOOL_SUBST_CTRL) {
		Debug_out(DEBUG, "Past length: %d\n", new->maxPastLength);
		Debug_out(DEBUG, "Past Distribution: \n");
		for (i = 0; i < new->maxPastLength; i++) {
			printf("\t%lg", new->pastDistribution[i]);
		}
		printf("\n");
		fflush(NULL);
	}

	if (new->tool == TOOL_CTRL) {
		Debug_out(DEBUG, "Absolute path of useHistoryFile: %s\n", new->useHistoryFile);
	}
	if (new->tool == TOOL_SUBST_CTRL) {
		Debug_out(DEBUG, "Energy Window length: %d\n", new->energyWindowSubst_l);
	}


	if (new->tariff == TARIFF_IBR) {
		Debug_out(DEBUG, "Absolute path of IBR pricePolicyFile: %s\n", new->pricePolicyFile);
		Debug_out(DEBUG, "IBR Tariff for BUY LOW: %lg\n", new->IBR[IBR_BUY_LOW]);
		Debug_out(DEBUG, "IBR Tariff for BUY HIGH: %lg\n", new->IBR[IBR_BUY_HIGH]);
		Debug_out(DEBUG, "IBR Tariff for SELL: %lg\n", new->IBR[IBR_SELL]);
	}
	else if (new->tariff == TARIFF_TOU) {
		Debug_out(DEBUG, "Absolute path of tariffsToUFile: %s\n", tariffsToUFile);
	}
	else if (new->tariff == TARIFF_FLAT) {
		Debug_out(DEBUG, "FLAT Tariff for BUY: %lg\n", new->FLAT[TOU_BUY]);
		Debug_out(DEBUG, "FLAT Tariff for SELL: %lg\n", new->FLAT[TOU_SELL]);
	}
	else if (new->tariff == TARIFF_SUBST) {
		Debug_out(DEBUG, "Absolute path of tariffsSubstFile: %s\n", new->tariffsSubstFile);
	}

	Debug_out(DEBUG, "Absolute path of loggingFile: %s\n", new->loggingFile);
	Debug_out(DEBUG, "Absolute path of batteryFile: %s\n", new->batteryFile);
	if (strcmp(new->phevFile, "") != 0) {
		Debug_out(DEBUG, "Absolute path of phevFile: %s\n", new->phevFile);
	}
	else {
		Debug_out(DEBUG, "There is NOT a PHEV plugged\n");
	}


	/* Open logging file */
	new->loggingFile_fd = fopen(new->loggingFile, "w");
	if (!new->loggingFile_fd) {
		perror("Problems opening logging file");
		exit(1);
	}
	if (new->tool == TOOL_SUBST_CTRL || new->tool == TOOL_SUBST_HISTORY) {
		fprintf(new->loggingFile_fd, "real_energy_kwh,energy_from_MILP_kwh,demand_kwh,battery_soc_kwh,charge_battery_kw,real_energy_price,applied_tariff");
	}
	else {
		fprintf(new->loggingFile_fd, "real_energy_grid_kwh,energy_grid_from_MILP_kwh,consumption_kwh,production_kwh,battery_soc_kwh,phev_soc_kwh,phev_ready_hours,charge_battery_kw,charge_phev_kw,phev_next_soc_kwh,real_energy_price,applied_tariff");
	}
	/* if price policy dump also max price policy  */
	if (new->tariff == TARIFF_IBR) {
		fprintf(new->loggingFile_fd, ",price_policy_max");
	}
	/* if there is a forecast based on heuristic, dump all forecast */
	if (new->tool == TOOL_CTRL) {
		for (i = 0; i < TIMESLOTS_AHEAD-1; i++) {
			fprintf(new->loggingFile_fd, ",predicted_%d_consumption_kwh,predicted_%d_production_kwh,predicted_%d_phev_soc_kwh", i+1, i+1, i+1);
		}
	}
	else if (new->tool == TOOL_SUBST_CTRL) {
		for (i = 0; i < TIMESLOTS_AHEAD-1; i++) {
			fprintf(new->loggingFile_fd, ",predicted_%d_demand_kwh", i+1);
		}
	}
	fprintf(new->loggingFile_fd, "\n");
	fflush(new->loggingFile_fd);


	/* Battery */
	if (new->tool != TOOL_HISTORY && new->tool != TOOL_SUBST_HISTORY) {
		new->battery = Battery_new(new->batteryFile);
		Debug_out(DEBUG, "Battery:\n");
		Debug_perform(DEBUG, Battery_dump(new->battery, stdout));
	}
	else {
		new->battery = NULL;
		Debug_out(DEBUG, "Discarding Battery (in simulation)\n");
	}

	/* PHEV */
	if (strcmp(new->phevFile, "") != 0) {
		new->PHEV = Battery_new(new->phevFile);
		Debug_out(DEBUG, "PHEV:\n");
		Debug_perform(DEBUG, Battery_dump(new->PHEV, stdout));
	}
	else {
		new->PHEV = NULL;
		Debug_out(DEBUG, "Setting PHEV to NULL\n");
	}

	/* ToU Tariffs reading */
	if (new->tariff == TARIFF_TOU) {
		EBR_readToUTariffs(new, tariffsToUFile);
		new->currentHour = -1;
	}

	/* Use real historical data, if need be */
	if (new->tool == TOOL_EVAL || new->tool == TOOL_HISTORY || new->tool == TOOL_CTRL) {
		new->useHistoryFile_fd = fopen(new->useHistoryFile, "r");
		if (!new->useHistoryFile_fd) {
			perror("Errors opening useHistoryFile");
			exit(1);
		}
		if (!fgets(line, MAX_BUF_LEN, new->useHistoryFile_fd)) {
			fprintf(stderr, "Problems reading from %s\n", new->useHistoryFile);
			exit(1);
		}
		if (line[strlen(line)-1] == '\n') {
			line[strlen(line)-1] = '\0';
		}
		Debug_out(DEBUG, "Profiles first line : %s", line);
	}

	/* Learn from past, if need be */
	if (new->tool == TOOL_CTRL || new->tool == TOOL_SUBST_CTRL) {
		for (i = 0; i < TIMESLOTS_IN_A_DAY; i++) {
			new->pastLength[i] = 0;
		}
		for (i = 0; i < FORECASTS_NUMBER; i++) {
			for (j = 0; j < TIMESLOTS_IN_A_DAY; j++) {
				new->past[i][j] = (double *) calloc(new->maxPastLength, sizeof(double));
				if (new->past[i][j] == NULL) {
					perror("Problems in allocating memory for past");
					exit(1);
				}
			}
		}
	}

	/* timeslot length */
	//	new->timeslotLength = CommandLineInterface_getIntOption(cliPtr, "timeslotLength");
	//	Debug_out(DEBUG, "Timeslot length for EBR = %d\n", new->timeslotLength);
	//	new->timeslotsInHour = 60/new->timeslotLength;
	//	Debug_out(DEBUG, "Number of EBR timeslots in a hour = %d\n", new->timeslotsInHour);
	new->timeslotLength = 60;
	Debug_out(DEBUG, "Timeslot length for EBR = %d\n", new->timeslotLength);
	new->timeslotsInHour = 1;
	Debug_out(DEBUG, "Number of EBR timeslots in a hour = %d\n", new->timeslotsInHour);

	/* hours ahead for forecasting */
	//	new->forecastHours  = CommandLineInterface_getDoubleOption(cliPtr, "forecastHours");
	//	Debug_out(DEBUG, "Number of horus for forecasting = %lg\n", new->forecastHours);
	//	new->timeslotsAhead = (int) ceil(new->forecastHours * new->timeslotsInHour);
	//	Debug_out(DEBUG, "Number of EBR timeslots for forecasting = %d\n", new->timeslotsAhead);
	new->forecastHours  = TIMESLOTS_AHEAD;
	Debug_out(DEBUG, "Number of hours for forecasting = %lg\n", new->forecastHours);
	new->timeslotsAhead = TIMESLOTS_AHEAD;
	Debug_out(DEBUG, "Number of EBR timeslots for forecasting = %d\n", new->timeslotsAhead);
	
	/* Price policy */
	if (new->tariff == TARIFF_IBR) {
		new->pricePolicyFileCurrentLine = 0;
		new->pricePolicy[PRICEPOLICY_MIN] = (double *) calloc(new->timeslotsAhead, sizeof(double));
		new->pricePolicy[PRICEPOLICY_MAX] = (double *) calloc(new->timeslotsAhead, sizeof(double));
		new->pricePolicyFile_fd = fopen(new->pricePolicyFile, "r");
		if (!new->pricePolicyFile_fd) {
			fprintf(stderr, "Error opening file %s\n", new->pricePolicyFile);
			exit(1);
		}
		if (!fgets(line, MAX_BUF_LEN, new->pricePolicyFile_fd)) {
			fprintf(stderr, "Problems reading from %s\n", new->pricePolicyFile);
			exit(1);
		}
		if (line[strlen(line)-1] == '\n') {
			line[strlen(line)-1] = '\0';
		}
		Debug_out(DEBUG, "Price policies first line : %s\n", line);
	}
	else if (new->tariff == TARIFF_SUBST) {
		new->tariffsSubstFile_fd = fopen(new->tariffsSubstFile, "r");
		if (!new->tariffsSubstFile_fd) {
			fprintf(stderr, "Error opening file %s\n", new->tariffsSubstFile);
			exit(1);
		}
		if (!fgets(line, MAX_BUF_LEN, new->tariffsSubstFile_fd)) {
			fprintf(stderr, "Problems reading from %s\n", new->tariffsSubstFile);
			exit(1);
		}
		if (line[strlen(line)-1] == '\n') {
			line[strlen(line)-1] = '\0';
		}
		Debug_out(DEBUG, "Substation Tariff first line : %s\n", line);
		new->tariffsSubstFile_started = 0;
	}

	Debug_out(DEBUG, "TIMESLOTS IN A DAY = %d\n", TIMESLOTS_IN_A_DAY);
	Debug_out(DEBUG, "TOLERABLE PHEV ABSENCE = %lg\n", TOLERABLE_PHEV_ABSENCE);
	Debug_out(DEBUG, "PHEV CHARGE FACTOR = %lg\n", PHEV_CHARGE);

	EBR_initPastTraining(new);

	free((char *)tariff);
	tariff = NULL;
	free((char *)tool);
	tool = NULL;
	free(pastDistribution);
	pastDistribution = NULL;

	return new;
}



void EBR_free(EBR **_this)
{
	Debug_assert(DEBUG, *_this != NULL, "*_this == NULL\n");
	int i, j;

	Properties_free(&(*_this)->configHECH);
	fclose((*_this)->loggingFile_fd);
	Battery_free(&(*_this)->battery);
	if ((*_this)->PHEV) {
		Battery_free(&(*_this)->PHEV);
	}
	if ((*_this)->tool == TOOL_EVAL || (*_this)->tool == TOOL_HISTORY || (*_this)->tool == TOOL_CTRL) {
		fclose((*_this)->useHistoryFile_fd);
	}
	if ((*_this)->tool == TOOL_CTRL || (*_this)->tool == TOOL_SUBST_CTRL) {
		for (i = 0; i < FORECASTS_NUMBER; i++) {
			for (j = 0; j < TIMESLOTS_IN_A_DAY; j++) {
				free((*_this)->past[i][j]);
				(*_this)->past[i][j] = NULL;
			}
		}
	}
	if ((*_this)->tool == TOOL_SUBST_CTRL) {
		free((*_this)->energyWindowSubst);
		(*_this)->energyWindowSubst = NULL;
	}
	if ((*_this)->tariff == TARIFF_IBR) {
		fclose((*_this)->pricePolicyFile_fd);
		for (i = 0; i < PRICEPOLICY_NUMBER; i++) {
			free((*_this)->pricePolicy[i]);
		}
	}
	else if ((*_this)->tariff == TARIFF_SUBST) {
		fclose((*_this)->tariffsSubstFile_fd);
	}
	free(*_this);
	*_this = NULL;
}




static void _EBR_readEnergyFile(EBR *_this, char * energyFile)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	static char line[MAX_BUF_LEN];

	FILE *energyFile_fd;
	char *token;

	energyFile_fd = fopen(energyFile, "r");
	if (!energyFile_fd) {
		fprintf(stderr, "Error opening file %s\n", energyFile);
		exit(1);
	}
	fgets(line, MAX_BUF_LEN, energyFile_fd);
	if (!fgets(line, MAX_BUF_LEN, energyFile_fd)) {
		fprintf(stderr, "Problems reading from %s\n", energyFile);
		exit(1);
	}
	if (line[strlen(line)-1] == '\n') {
		line[strlen(line)-1] = '\0';
	}

	/* max energy */
	token = strtok(line, ",");
	if (!token) {
		perror("ERROR in tokenize");
		exit(1);
	}
	if (StringUtils_toDouble(token, &(_this->maxEnergy)) != 0) {
		Debug_out(DEBUG, "Problems reading max contractual energy, value %s not accepted\n", token);
		exit(1);
	}
	/* min energy */
	token = strtok(NULL, ",");
	if (!token) {
		perror("ERROR in tokenize");
		exit(1);
	}
	if (StringUtils_toDouble(token, &(_this->minEnergy)) != 0) {
		Debug_out(DEBUG, "Problems reading min contractual energy, value %s not accepted\n", token);
		exit(1);
	}

	fclose(energyFile_fd);
}



static void _EBR_pastDistribution(EBR *_this, int maxPastLength, char *pastDistribution)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	Debug_assert(DEBUG, _this->tool == TOOL_CTRL || _this->tool == TOOL_SUBST_CTRL, "!(_this->tool == TOOL_CTRL || _this->tool == TOOL_SUBST_CTRL)\n");
	int i;
	char * token;
	double sum;
	_this->pastDistribution = (double *) calloc(maxPastLength, sizeof(double));
	/* fill pastDistribution */
	for (i = 0; i < maxPastLength; i++) {
		token = strtok(i==0? pastDistribution : NULL, ":");
		if (!token) {
			perror("ERROR in tokenize pastDistribution");
			exit(1);
		}
		if (StringUtils_toDouble(token, &(_this->pastDistribution[i])) != 0) {
			Debug_out(DEBUG, "Problems reading --pastDistribution, value %s not accepted\n", token);
			exit(1);
		}
	}
	/* check */
	printf("Past Distribution:");
	sum = 0;
	for (i = 0; i < maxPastLength; i++) {
		printf(" %lg", _this->pastDistribution[i]);
		sum += _this->pastDistribution[i];
	}
	printf("\n");
	if (sum != 1) {
		Debug_out(DEBUG, "--pastDistribution %s elements do not sum to 1, they sum to %lg instead\n", pastDistribution, sum);
	}
}


