#include <HECH.h>
#include <Debug.h>
#include <StringUtils.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <netdb.h>


#define BUFSIZE 1024

/* To make it working on Linux we need this fake definition... */
#define h_addr h_addr_list[0]


#define DEBUG "HECH"



struct HECH {
	char hostName[BUFSIZE];
	int measurementsPort;
	int commandsPort;
	int controlSignalStop;

	int sockfdMeas;
	int sockfdCmds;

	/* fields added for evaluation */
	long actualControlSignal;
};



HECH * HECH_new(const char * hostName, int measurementsPort, int commandsPort, int controlSignalStop)
{
	struct sockaddr_in serv_addr_meas;
	struct sockaddr_in serv_addr_cmds;
	struct hostent *server_meas;
	struct hostent *server_cmds;

	HECH *new = (HECH *) calloc(1, sizeof(HECH));
	sprintf(new->hostName, "%s", hostName);
	new->measurementsPort = measurementsPort;
	new->commandsPort = commandsPort;
	new->controlSignalStop = controlSignalStop;
	new->actualControlSignal = 0;

	/** Connects to House server */

	/* Measurements */
	new->sockfdMeas = socket(AF_INET, SOCK_STREAM, 0);
	if (new->sockfdMeas < 0) {
		Debug_out(DEBUG, "ERROR opening socket for requesting measurements\n");
		return NULL;
	}
	server_meas = gethostbyname(hostName);
	if (server_meas == NULL) {
		Debug_out(DEBUG, "ERROR, no such host %s\n", hostName);
		return NULL;
	}
	bzero((char *) &serv_addr_meas, sizeof(serv_addr_meas));
	serv_addr_meas.sin_family = AF_INET;
	bcopy((char *)server_meas->h_addr,
		  (char *)&serv_addr_meas.sin_addr.s_addr,
		  server_meas->h_length);
	serv_addr_meas.sin_port = htons(new->measurementsPort);
	/* Now connect to the server */
	if (connect(new->sockfdMeas, (struct sockaddr *) &serv_addr_meas, sizeof(serv_addr_meas)) < 0) {
		Debug_out(DEBUG, "ERROR connecting to House server for requesting measurements at %s:%d\n", hostName, new->measurementsPort);
		return NULL;
	}
	Debug_out(DEBUG, "Connected to %s:%d for measurement requests\n", hostName, new->measurementsPort);

	/* Commands */
	new->sockfdCmds = socket(AF_INET, SOCK_STREAM, 0);
	if (new->sockfdCmds < 0) {
		Debug_out(DEBUG, "ERROR opening socket for commands\n");
		return NULL;
	}
	server_cmds = gethostbyname(hostName);
	if (server_cmds == NULL) {
		Debug_out(DEBUG, "ERROR, no such host %s\n", hostName);
		return NULL;
	}
	bzero((char *) &serv_addr_cmds, sizeof(serv_addr_cmds));
	serv_addr_cmds.sin_family = AF_INET;
	bcopy((char *)server_cmds->h_addr,
		  (char *)&serv_addr_cmds.sin_addr.s_addr,
		  server_cmds->h_length);
	serv_addr_cmds.sin_port = htons(new->commandsPort);
	/* Now connect to the server */
	if (connect(new->sockfdCmds, (struct sockaddr *) &serv_addr_cmds, sizeof(serv_addr_cmds)) < 0) {
		Debug_out(DEBUG, "ERROR connecting to House server for commands at %s:%d\n", hostName, new->commandsPort);
		return NULL;
	}
	Debug_out(DEBUG, "Connected to %s:%d for commands\n", hostName, new->commandsPort);

	return new;
}


void HECH_free(HECH **_this)
{
	Debug_assert(DEBUG, *_this != NULL, "*_this == NULL\n");
	/* Disconnects from House server */
	HECH * this = *_this;
	Debug_out(DEBUG, "Closing connection with %s:%d\n", this->hostName, this->measurementsPort);
	Debug_out(DEBUG, "Closing connection with %s:%d\n", this->hostName, this->commandsPort);
	close(this->sockfdMeas);
	close(this->sockfdCmds);
	free(this);
	*_this = NULL;
}


int HECH_getControlStopSignal(HECH *_this)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	return _this->controlSignalStop;
}



int HECH_requestMeasurements(HECH *_this, double *meas)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	Debug_assert(DEBUG, meas != NULL, "meas == NULL\n");

	int n, i;
	static long int readControl;

	/* Updates control signal */
	_this->actualControlSignal = _this->actualControlSignal + 1;

	/* Send message to the server */
	n = send(_this->sockfdMeas, (char *)&(_this->actualControlSignal), sizeof(long int), 0);
	if (n <= 0) {
		Debug_out(DEBUG, "ERROR sending measurements request '%ld'\n", _this->actualControlSignal);
		return -2;
	}
	Debug_out(DEBUG, "Sent message '%ld'\n", _this->actualControlSignal);

	/* Read server response */
	n = recv(_this->sockfdMeas, (char *)&readControl, sizeof(long int), 0);
	if (n < 0) {
		Debug_out(DEBUG, "ERROR reading from socket while receiving measurements\n");
		return -2;
	}
	else if (n == 0) {
		Debug_out(DEBUG, "ERROR server closed its connection\n");
		return -2;
	}
	/* stop if need be */
	if (readControl == _this->controlSignalStop || readControl == -2) {
		return readControl;
	}
	Debug_out(DEBUG, "readControl = %ld\n", readControl);

	/* Fill parameter meas with answer from server */
	for (i = 0; i < MEAS_NUMBER; i++) {
		n = recv(_this->sockfdMeas, (char *)&meas[i], sizeof(double), 0);
		if (n < 0) {
			Debug_out(DEBUG, "ERROR reading from socket while receiving measurements\n");
			return -2;
		}
		else if (n == 0) {
			Debug_out(DEBUG, "ERROR server closed its connection\n");
			return -2;
		}
		Debug_out(DEBUG, "meas[%d] = %lg\n", i, meas[i]);
	}

	return readControl;
}


int HECH_sendCommands(HECH *_this, double *commands)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	Debug_assert(DEBUG, commands != NULL, "commands == NULL\n");

	int n, i;
	long int readControl = 0;

	/* Send message to the server */
	n = send(_this->sockfdCmds, (char *)&(_this->actualControlSignal), sizeof(long int), 0);
	if (n <= 0) {
		Debug_out(DEBUG, "ERROR sending commands '%ld'\n", _this->actualControlSignal);
		return -2;
	}
	for (i = 0; i < CMDS_NUMBER; i++) {
		n = send(_this->sockfdCmds, (char *)&commands[i], sizeof(double), 0);
		if (n <= 0) {
			Debug_out(DEBUG, "ERROR sending commands '%lg'\n", commands[i]);
			return -2;
		}
	}
	Debug_out(DEBUG, "Sent command '%ld %lg %lg'\n", _this->actualControlSignal, commands[CMDS_BATTERY], commands[CMDS_PHEV]);

	/* Read server response */
	n = recv(_this->sockfdCmds, (char *)&readControl, sizeof(long int), 0);
	if (n < 0) {
		Debug_out(DEBUG, "ERROR reading commands ack\n");
		return -2;
	}
	else if (n == 0) {
		Debug_out(DEBUG, "ERROR server closed its connection\n");
		return -2;
	}
	Debug_out(DEBUG, "Read message '%ld'\n", readControl);

	return readControl;
}



void HECH_dumpInfo(HECH *_this, FILE *stream)
{
	Debug_assert(DEBUG, _this != NULL, "*_this == NULL\n");

	fprintf(stream, "==== HECH Information ==== (%p)\n", _this);
	fprintf(stream, "host name:\t%s\n", _this->hostName);
	fprintf(stream, "port for measurements requests:\t%d\n", _this->measurementsPort);
	fprintf(stream, "port for commands:\t%d\n", _this->commandsPort);
	fprintf(stream, "control signal stop:\t%d\n", _this->controlSignalStop);
	fprintf(stream, "==========================\n");
	fflush(stream);
}

