#include <House.h>
#include <Properties.h>
#include <Debug.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>


#define DEBUG "House Functions"

#define BACKLOG 10



static int _House_openSocketAtPort(int port);
static int _House_startServer(House *_this, int port, int (*serverFnc)(House *, int, int));

static int _House_serverProducer(House *_this, int pid, int sock);
static int _House_serverMeasReqsConsumer(House *_this, int pid, int sock);
static int _House_serverCmdsConsumer(House *_this, int pid, int sock);

static void *_House_get_in_addr(struct sockaddr *sa);


Properties * House_readConfigurationFile(const char *configFile)
{
	return Properties_newFromFile(configFile, 1024, 1024);
}


House * House_new(Properties *conf)
{
	House * new = (House *) calloc(1, sizeof(House));

	Properties_getInt(conf, "PRODUCER_LISTEN_PORT", &(new->producerPort));
	Properties_getInt(conf, "CONSUMER_LISTEN_MEAS_PORT", &(new->consumerMeasurementsPort));
	Properties_getInt(conf, "CONSUMER_LISTEN_CMDS_PORT", &(new->consumerCommandsPort));

	return new;
}


void House_free(House ** _this)
{
	Debug_assert(DEBUG, *_this != NULL, "*_this == NULL\n");
	free(*_this);
	*_this = NULL;
}





void House_startServers(House *_this)
{
	int res;

	Debug_out(DEBUG, "Starting server for Producers\n");
	res = _House_startServer(_this, _this->producerPort, &_House_serverProducer);
	if (res < 0) {
		fprintf(stderr, "Problems starting server for Producers\n");
		return;
	}
	Debug_out(DEBUG, "Server for Producers PID = %d\n", res);

	Debug_out(DEBUG, "Starting server for measurements reqs from Consumers\n");
	res = _House_startServer(_this, _this->consumerMeasurementsPort, &_House_serverMeasReqsConsumer);
	if (res < 0) {
		fprintf(stderr, "Problems starting server for measurements reqs from Consumers\n");
		return;
	}
	Debug_out(DEBUG, "Server for measurements reqs from Consumers PID = %d\n", res);

	Debug_out(DEBUG, "Starting server for commands from Consumers\n");
	res = _House_startServer(_this, _this->consumerCommandsPort, &_House_serverCmdsConsumer);
	if (res < 0) {
		fprintf(stderr, "Problems starting server for commands from Consumers\n");
		return;
	}
	Debug_out(DEBUG, "Server for commands from Consumers PID = %d\n", res);
}



void House_waitServers(House *_this)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");

	/* Waits for all children to terminate */
	while(wait(NULL)>0);

}





static int _House_startServer(House *_this, int port, int (*serverFnc)(House *, int, int))
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	pid_t childProcess;

	childProcess = fork();

	/* Child code */
	if (childProcess == 0) {
		/* child is server always accepting requests */
		int fncres;
		pid_t thispid = getpid();

		/* Open server for measurements requested by Consumers */
		int sock = _House_openSocketAtPort(port);
		if (sock == -1) {
			exit(-1);
		}
		Debug_out(DEBUG, "[%d] Opened server at port %d\n", thispid, port);

		/* accepts until control signal STOP is read */
		fncres = serverFnc(_this, thispid, sock);

		/* server shuts down */
		close(sock);
		Debug_out(DEBUG, "[%d] Closed server at port %d\n", thispid, port);
		exit(0);
	}
	else if (childProcess < 0) { /* Error in fork() */
		fprintf(stderr, "Error in creating server\n");
		return -1;
	}

	return childProcess;
}





/* Returns socket fd if everything is OK, -1 otherwise */
static int _House_openSocketAtPort(int port)
{
	struct sockaddr_in sin;

	/* Creates socket */
	int sock = socket(AF_INET, SOCK_STREAM, 0);
	if(sock == -1) {
		fprintf(stderr, "Error opening socket\n");
		return -1;
	}
	sin.sin_port = htons(port);
	sin.sin_addr.s_addr = 0;
	sin.sin_addr.s_addr = INADDR_ANY;
	sin.sin_family = AF_INET;
	/* Binds socket */
	if(bind(sock, (struct sockaddr *)&sin,sizeof(struct sockaddr_in) ) == -1) {
		fprintf(stderr, "Error binding socket at port %d\n", port);
		return -1;
	}
	/* Listens on socket */
	if (listen(sock, BACKLOG) == -1) {
		fprintf(stderr, "Error listening on soket at port %d\n", port);
		exit(1);
	}

	return sock;
}



static int _House_serverProducer(House *_this, int pid, int sock)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");

	struct sockaddr_storage their_addr; // connector's address information
	socklen_t sin_size;
	int new_fd;
	char s[INET6_ADDRSTRLEN];

	Debug_out(DEBUG, "[%d] Accepting requests\n", pid);

	while (1) {
		sin_size = sizeof(their_addr);

		new_fd = accept(sock, (struct sockaddr *)&their_addr, &sin_size);

		if (new_fd == -1) {
			fprintf(stderr, "[%d] ERROR accepting", pid);
			continue;
		}

		inet_ntop(their_addr.ss_family,
				  _House_get_in_addr((struct sockaddr *)&their_addr),
				  s, sizeof(s));
		Debug_out(DEBUG, "[%d] Connection from %s\n", pid, s);

		if (!fork()) { // this is the child process
			close(sock); // child doesn't need the listener

			/* DO SOMETHING HERE */

			close(new_fd);
			exit(0);
		}
		close(new_fd);  // parent doesn't need this descriptor
	}

	return 0;
}


static int _House_serverMeasReqsConsumer(House *_this, int pid, int sock)
{
	return 0;
}


static int _House_serverCmdsConsumer(House *_this, int pid, int sock)
{
	return 0;
}

// get sockaddr, IPv4 or IPv6:
void * _House_get_in_addr(struct sockaddr *sa)
{
	if (sa->sa_family == AF_INET) {
		return &(((struct sockaddr_in*)sa)->sin_addr);
	}

	return &(((struct sockaddr_in6*)sa)->sin6_addr);
}
