#define _GNU_SOURCE

#include <House.h>
#include <Properties.h>
#include <CommandLineInterface.h>
#include <Debug.h>


#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <strings.h>


#define DEBUG "House Main"


int main(int argc, char **argv)
{
	Properties * conf = NULL;
	CommandLineInterface * cliPtr = NULL;
	const char * configFile;
	char profilesFile[PATH_MAX];
	char loggingFile[PATH_MAX];
	char batteryFile[PATH_MAX];
	const char * phev;
	char phevFile[PATH_MAX];
	int onlySimulateProfiles;
	int isSubstation;

	/** CLI */
	cliPtr = CommandLineInterface_new(argv[0], "House simulator (LabVIEW in IMDEA labs)");
	/* Add CLI options */
	CommandLineInterface_addStringOption_withDefault(cliPtr, "configFile", "House_conf.ini", "House Configuration file");
	CommandLineInterface_addStringOption_withDefault(cliPtr, "profilesFile", "profiles.csv", "Profiles file name, in CSV format");
	CommandLineInterface_addStringOption_withDefault(cliPtr, "loggingFile", "LabVIEWloggedProfiles.csv", "Output file name for real profiles logging file, in CSV format");
	CommandLineInterface_addStringOption_withDefault(cliPtr, "batteryFile", "battery.csv", "File with properties of battery, in CSV format");
	CommandLineInterface_addStringOption_withDefault(cliPtr, "phevFile", NULL, "File with properties of PHEV, in CSV format");
	CommandLineInterface_addFlag(cliPtr, "onlySimulateProfiles", "Only perform a simulation of profiles data in file specified by --useHistoryFile");

	CommandLineInterface_addFlag(cliPtr, "isSubstation", "Act as server in a Substation instead of a House (this implies that there is no PHEV, so no --phevFile must be passed)");

	/* Parse and get CLI options */
	CommandLineInterface_processArgs(cliPtr, argc, argv);
	configFile = CommandLineInterface_getStringOption(cliPtr, "configFile");
	realpath(CommandLineInterface_getStringOption(cliPtr, "profilesFile"), profilesFile);
	realpath(CommandLineInterface_getStringOption(cliPtr, "loggingFile"), loggingFile);
	realpath(CommandLineInterface_getStringOption(cliPtr, "batteryFile"), batteryFile);
	isSubstation = CommandLineInterface_getFlag(cliPtr, "isSubstation");
	if (!isSubstation) {
		phev = CommandLineInterface_getStringOption(cliPtr, "phevFile");
		if (phev) {
			realpath(phev, phevFile);
		}
		else {
			bzero(phevFile, PATH_MAX);
		}
	}
	else {
		phev = 0;
	}
	onlySimulateProfiles = CommandLineInterface_getFlag(cliPtr, "onlySimulateProfiles");
	Debug_out(DEBUG, "Command Line Interface:\n");
	Debug_perform(DEBUG, CommandLineInterface_fprint_args(cliPtr, stdout));
	Debug_out(DEBUG, "Absolute path of profilesFile: %s\n", profilesFile);
	Debug_out(DEBUG, "Absolute path of loggingFile: %s\n", loggingFile);
	Debug_out(DEBUG, "Absolute path of batteryFile: %s\n", batteryFile);
	if (phev) {
		Debug_out(DEBUG, "Absolute path of phevFile: %s\n", phevFile);
	}
	else {
		Debug_out(DEBUG, "No PHEV\n");
	}
	if (onlySimulateProfiles) {
		Debug_out(DEBUG, "\nONLY SIMULATING HISTORICAL DATA\n");
	}

	/* Reads parameters from configuration file */
	conf = Properties_newFromFile(configFile, 1024, 1024);
	if (!conf) {
		if (cliPtr != NULL) CommandLineInterface_free(&cliPtr);
		return 1;
	}
	Debug_out(DEBUG, "Configuration file %s:\n", configFile);
	Debug_perform(DEBUG, Properties_fprint(conf, stdout));


	House *LabVIEW = House_new(conf, profilesFile, loggingFile, batteryFile, phevFile, onlySimulateProfiles, isSubstation);
	if (LabVIEW == NULL) {
		fprintf(stderr, "Problems opening sockets\n");
		exit(1);
	}

	/** Starts servers */
	House_startServers(LabVIEW);

	/** Waits until all servers receive control signal STOP (can go on forever) */
	House_waitServers(LabVIEW);

	/** Exit */
	House_free(&LabVIEW);
	if (conf != NULL) Properties_free(&conf);
	if (cliPtr != NULL) CommandLineInterface_free(&cliPtr);

	return 0;
}
