#ifndef __ARRAY__
#define __ARRAY__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <Debug.h>


typedef struct Array Array;

/** Creates a new Array whose initial capacity is 'capacity'. The array
	will contain elements whose size is 'elemSize'. 
	When full, it will automatically enlarge by 'capIncrement'.
	
	Function fprintElem prints on a file stream a textual representation
	of an element. It receives as input a pointer to an element of the array.
	
	Function fEqualsElem checks whether two elements are equal.
	It receives as input pointers to two elements of the array.
*/
Array* Array_new(int capacity, size_t elemSize, int capIncrement, 
	void (*fprintElem)(const void* const elem_p, FILE*),
	int (*fEqualsElem)(const void* const e1_p, const void* const e2_p) );
	
/** Creates an array whose elements are int */
Array* Array_newInt(int capacity, int capIncrement);

/** Creates an array whose elements are float */
Array* Array_newFloat(int capacity, int capIncrement);

/** Creates an array whose elements are double */
Array* Array_newDouble(int capacity, int capIncrement);

/** Creates an array whose elements are char* */
Array* Array_newString(int capacity, int capIncrement);

/** Creates an array whose elements are long */
Array* Array_newLong(int capacity, int capIncrement);

/** Creates an array by wrapping malloc-/calloc-ed C array 'buffer', whose 'length' elements 
	are assumed to have size 'elemSize'.
	Upon Array_free(), also buffer will be freed.
**/
Array* const Array_new_wrap(int length, size_t elemSize, 
		void* buffer,
		void (*fprintElem)(const void* const elem_p, FILE*),
		int (*fEqualsElem)(const void* const e1_p, const void* const e2_p)   );

Array* Array_clone(const Array* const _this);


size_t Array_elemSize(const Array* const _this);


void* Array_get(const Array* const _this, int i, void* const value);
void Array_set(Array* const _this, int i, const void* const value); 

void* Array_getFirst(const Array* const _this, void* const value);
void* Array_getLast(const Array* const _this, void* const value);

/** Returns a read-only pointer to the i-th element of this array.
	This pointer is guaranteed to be valid only until the next invocation
	of _add(), _push(), or any method which writes into this Array 
	(as they might trigger the capacity adjustment algorithm which could
	in turn trigger a realloc()).
	
	Contract: Array_elemPtr(array, i) == & (Array_as_array(array)[i])
**/
const void* Array_elemPtr(const Array* const _this, int i);

const void* const Array_as_C_array(const Array* const _this);

// Methods to handle this array as a stack
void* Array_pop(Array* const _this, void* const value);
void Array_push(Array* const _this, const void* const valueP);
void* Array_top(Array* const _this, void* const value);

// Methods to handle this array as a queue
void Array_enqueue(Array* const _this, const void* const valueP);
void* Array_dequeue(Array* const _this, void* const value);

void Array_add(Array* const _this, const void* const value);

void Array_swap(Array* _this, int i, int j);
void Array_trunc(Array* const _this, int len);

void Array_clear(Array* const _this);

int Array_length(const Array* const _this);

void Array_copy(const Array* const this, Array* dest);

void (*Array_fprintElem(const Array* const _this))(const void* const, FILE*);


int Array_equals(const Array* const this, const Array* const other);

void Array_free(Array** thisP);

void Array_fprint(const void* const _this, FILE* f);

#endif

