#ifndef __DEBUG_H__
#define __DEBUG_H__

/**
	The Debug package allows developers to easily program with assertions. This programming style:
	 	- reduces the risk of introducing bugs, and
		- allows the programmer to find a bug much easier
	
	For each program unit (e.g., a file defining a data-structure well as its related functions), 
	do the following:
		- #define DEBUG "DataStructure_Name"  // a string identifying 
			the file/data-structure, e.g., its name. This is called a "Debug category"
			(the macro name, e.g., DEBUG, can be whatever.)
		
		for each function do the following:
		
		int myFunction(MyDataStructure* this, int a, int b) {
			// Check pre-conditions: this must be != NULL
			Debug_assert(DEBUG, this != NULL, "this == NULL\n"); 
			
			// Check pre-conditions: a must be less than b (note the extra arguments, a-la printf)
			Debug_assert(DEBUG, a < b, "a = %d >= b = %d\n", a, b); 
			...
			a += b;
			// Debug messages
			Debug_out(DEBUG, "Now a = %d\n", a);
			...
			Debug_assert(DEBUG, a == myOtherWayOfComputingA(), "Computation of a is wrong\n");
			
			// Executes a function/instruction
			Debug_perform(DEBUG, printMyDataStructure() );
			
			int myarray[10] = {0}
			// populate myarray
			...
			
			// Print-out a debug message showing the content of array 'myarray'
			Debug_outArray(DEBUG, 
				"int myarray holding my numbers", // header to be printed
				"%d", // printf format to print out each array entry
				myarray, // array var
				10 // array length
			);
			
		}
		
		Assertions and debug messages are enabled by default for all categories, 
		but can be disabled in the main() function on category basis:
		
		Debug_setDefaultAssertions(boolean value); // enables/disables default value for assertion checking 
		
		Debug_setDefaultOut(boolean value); // enables/disables default debug output (Debug_out)
		and function execution (Debug_perform)
		
		Debug_enableOut(<category>); // explicitly enables debug output (Debug_out) and function execution
		(Debug_perform) for category <category>
		
		Debug_disableOut(<category>); // explicitly disables debug output and function execution 
		for category <category>

		Debug_enableAssertions(<category>); // explicitly enables assertion checking (Debug_assert) 
		for category <category>
		
		Debug_disableAssertions(<category>); // explicitly disables assertion checking 
		for category <category>


	
		Debug output and assertion checking for the special category "ALWAYS" 
		(in #define constants DEBUG_ALWAYS and DEBUG_ALL (DEBUG_ALL is deprecated)) are ALWAYS enabled.
		
		To disable all assertion checking at once (for, e.g., production code), simply compile
		with -DNDEBUG 
		
		Assertion checking and debug output are encoded as macros, so if -DNDEBUG is enabled during 
		compilation, there will be NO OVERHEAD in computation. All Debug_out, Debug_perform and Debug_assert
		are simply IGNORED by the compiler.
		
		Before terminating function main(), you should call Debug_freeAll() to deallocate the memory used
		by the Debug package.
*/


#include <assert.h>
#include <stdio.h>
#include <time.h>

#include <Properties.h>

#include <Timer.h>


#define DEBUG_ALWAYS "ALWAYS"
#define DEBUG_ALL "ALWAYS" // for back-ward compatibility, to be removed
#define CATEGORY_MAX_LENGTH 1024


void Debug_setDefaultOut(int d);
void Debug_setDefaultAssertions(int d);

void Debug_enableOut(const char* const id);
void Debug_enableAssertions(const char* const id);
void Debug_enable(const char* const id);
void Debug_disableOut(const char* const id);
void Debug_disableAssertions(const char* const id);
void Debug_disable(const char* const id);

unsigned long long Debug_timestamp_millisec();
#define TIMESTAMP (Debug_timestamp_millisec())

#define _OUT 1
#define _ASSERTIONS 2
int Debug_isEnabled(const char* const id, int WHAT);

void Debug_freeAll();


#ifndef NDEBUG
	#ifndef NDEBUG_OUT
		#define Debug_out(id, ...) \
			if (Debug_isEnabled(id, _OUT)) do { \
				fprintf(stderr, "[timestamp=%llu] %s:%d: ", TIMESTAMP, __FILE__, __LINE__); \
				fprintf(stderr, __VA_ARGS__); \
			} while(0)

		#define Debug_perform(id, f) \
			if (Debug_isEnabled(id, _OUT)) do { \
				f; \
			} while(0)	
			
		#define Debug_outArray(id, name, format, array, length) \
			if (Debug_isEnabled(id, _OUT)) do { \
				Debug_out(id, "Array %s (length %d):\n", name, length); \
				for(int i=0; i<length; i++) { \
					Debug_out(id, "  [%d] --> ", i); \
					fprintf(stderr, format, array[i]); \
					fprintf(stderr, "\n"); \
				} \
				Debug_out(id, "end\n");\
			} while(0)
		
	#else // case NDEBUG_OUT defined
		#define Debug_out(...)((void) 0)
		#define Debug_perform(id, f)((void)0)
		#define Debug_outArray(id, name, format, array, length)((void) 0)

	#endif


	#define Debug_assert(id, cond, ...) \
		if (Debug_isEnabled(id, _ASSERTIONS)) do { \
			if (! (cond) ) { \
				fprintf(stderr, "[timestamp=%llu] %s:%d: ", TIMESTAMP, __FILE__, __LINE__); \
				fprintf(stderr, __VA_ARGS__); \
			} \
			assert(cond); \
		} while(0)

#else // case NDEBUG defined

	#define Debug_assert(id, cond, ...)((void)0)

	#define Debug_out(...)((void) 0)
	#define Debug_outArray(id, name, format, array, length)((void) 0)
	#define Debug_perform(id, f)((void)0)

#endif


#endif
