#define _GNU_SOURCE

#include <Battery.h>
#include <Debug.h>
#include <StringUtils.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>


#define DEBUG "Battery"


struct Battery {

	double stateOfCharge;

	/* Parameters */

	double capacity;

	double chargeSpeed;

	double dischargeSpeed;

	double chargeFactor;

	double dischargeFactor;

};



Battery *Battery_new(const char *filename)
{
	Debug_assert(DEBUG, filename != NULL, "filename == NULL\n");
	char * line = NULL;
	size_t len = 0;
	ssize_t read;
	char * token;
	FILE * BatteryFile_fd;
	Battery *new = (Battery *) calloc(1, sizeof(Battery));

	BatteryFile_fd = fopen(filename, "r");
	if (!BatteryFile_fd) {
		fprintf(stderr, "Error opening Battery capacity file %s\n", filename);
		perror("");
		exit(1);
	}

	read = getline(&line, &len, BatteryFile_fd); // first line is name (CSV format)
	read = getline(&line, &len, BatteryFile_fd);
	if (read == -1) {
		Debug_out(DEBUG, "Problems reading from %s\n", filename);
		perror("");
		exit(1);
	}

	/* capacity */
	token = strtok(line, ",");
	if (!token) {
		Debug_out(DEBUG, "Problems tokenizing '%s' with comma separator\n", line);
		return NULL;
	}
	if (StringUtils_toDouble(token, &(new->capacity)) != 0) {
		Debug_out(DEBUG, "Value %s not a valid double\n", token);
		return NULL;
	}

	/* charge speed */
	token = strtok(NULL, ",");
	if (!token) {
		Debug_out(DEBUG, "Problems tokenizing '%s' with comma separator\n", line);
		return NULL;
	}
	if (StringUtils_toDouble(token, &(new->chargeSpeed)) != 0) {
		Debug_out(DEBUG, "Value %s not a valid double\n", token);
		return NULL;
	}

	/* discharge speed */
	token = strtok(NULL, ",");
	if (!token) {
		Debug_out(DEBUG, "Problems tokenizing '%s' with comma separator\n", line);
		return NULL;
	}
	if (StringUtils_toDouble(token, &(new->dischargeSpeed)) != 0) {
		Debug_out(DEBUG, "Value %s not a valid double\n", token);
		return NULL;
	}

	/* charge factor */
	token = strtok(NULL, ",");
	if (!token) {
		Debug_out(DEBUG, "Problems tokenizing '%s' with comma separator\n", line);
		return NULL;
	}
	if (StringUtils_toDouble(token, &(new->chargeFactor)) != 0) {
		Debug_out(DEBUG, "Value %s not a valid double\n", token);
		return NULL;
	}

	/* discharge factor */
	token = strtok(NULL, ",");
	if (!token) {
		Debug_out(DEBUG, "Problems tokenizing '%s' with comma separator\n", line);
		return NULL;
	}
	if (StringUtils_toDouble(token, &(new->dischargeFactor)) != 0) {
		Debug_out(DEBUG, "Value %s not a valid double\n", token);
		return NULL;
	}

	fclose(BatteryFile_fd);
	return new;
}


void Battery_free(Battery **_this)
{
	Debug_assert(DEBUG, *_this != NULL, "*_this == NULL\n");
	free(*_this);
	*_this = NULL;
}


void Battery_dump(Battery *_this, FILE *stream)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	fprintf(stream, "- state of charge : %lg\n", _this->stateOfCharge);
	fprintf(stream, "- capacity : %lg\n", _this->capacity);
	fprintf(stream, "- charge speed : %lg\n", _this->chargeSpeed);
	fprintf(stream, "- discharge speed : %lg\n", _this->dischargeSpeed);
	fprintf(stream, "- charge factor : %lg\n", _this->chargeFactor);
	fprintf(stream, "- discharge factor : %lg\n", _this->dischargeFactor);
	fflush(NULL);
}



void Battery_setStateOfCharge(Battery *_this, double SoC)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	_this->stateOfCharge = SoC;
}


double Battery_getStateOfCharge(Battery *_this)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	return _this->stateOfCharge;
}


double Battery_getCapacity(Battery *_this)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	return _this->capacity;
}



double Battery_getChargeSpeed(Battery *_this)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	return _this->chargeSpeed;
}



double Battery_getDischargeSpeed(Battery *_this)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	return _this->dischargeSpeed;
}



double Battery_getChargeFactor(Battery *_this)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	return _this->chargeFactor;
}



double Battery_getDischargeFactor(Battery *_this)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	return _this->dischargeFactor;
}



void Battery_actuate(Battery *_this, double power, double hours)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	Debug_assert(DEBUG, -(_this->dischargeSpeed) <= power && power <= _this->chargeSpeed, "! (-(_this->dischargeSpeed) <= power && power <= _this->chargeSpeed))\n");

	double factor = power > 0? _this->chargeFactor : 1;

	_this->stateOfCharge = _this->stateOfCharge + hours * factor * power;

	/* Saturation, assuming an intelligent battery controller */
	if (_this->stateOfCharge > _this->capacity) {
		_this->stateOfCharge = _this->capacity;
	}
	else if (_this->stateOfCharge < 0) {
		_this->stateOfCharge = 0;
	}
}



void Battery_charge(Battery *_this, double power, double hours)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	Debug_assert(DEBUG, 0 < power && power <= _this->chargeSpeed, "! (0 < power && power <= _this->chargeSpeed)\n");
	double SoC;

	SoC = _this->stateOfCharge + hours * _this->chargeFactor * power;

	/* assuming an intelligent battery controller */
	if (SoC > _this->capacity) {
		SoC = _this->capacity;
	}

	_this->stateOfCharge = SoC;
}


void Battery_discharge(Battery *_this, double power, double hours)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	Debug_assert(DEBUG, 0 < power && power <= _this->dischargeSpeed, "! (0 < power && power <= _this->chargeSpeed)\n");
	double SoC;

	SoC = _this->stateOfCharge - hours * _this->dischargeFactor * power;

	/* assuming an intelligent battery controller */
	if (SoC < 0) {
		SoC = 0;
	}

	_this->stateOfCharge = SoC;
}


