#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <Debug.h>
#include <HashTable.h>
#include <StringUtils.h>

#include <Properties.h>

#define LINE_MAX_LEN 1024*1024

#define DEBUG_P "Properties"
#define INDENT_STRING ("  ")


int Properties_fprint_withIndent(const Properties* const this, FILE* f, int level);

// Properties_Content_Type methods
Properties_Content_Type _Properties_Content_Type_parse(const char* const t) {
	Properties_Content_Type result = STRING;
	if (StringUtils_equalsIgnoreCase(t, "int")) {
		result = INT;
	}
	else if (StringUtils_equalsIgnoreCase(t, "ulong")) {
		result = ULONG;
	}
	else if (StringUtils_equalsIgnoreCase(t, "long")) {
		result = LONG;
	}	
	else if (StringUtils_equalsIgnoreCase(t, "double")) {
		result = DOUBLE;
	}
	else if (StringUtils_equalsIgnoreCase(t, "boolean") || 
		StringUtils_equalsIgnoreCase(t, "bool")) {
			result = BOOLEAN;
	} 
	else if (StringUtils_equalsIgnoreCase(t, "properties")) {
			result = PROPERTIES;
	}
	return result;
}
const char* const Properties_Content_Type_toString(Properties_Content_Type type) {
	switch (type) {
		case INT : return "int";
		case ULONG: return "ulong";
		case LONG: return "long";
		case BOOLEAN : return "boolean";
		case DOUBLE : return "double";
		case STRING : return "string";
		case PROPERTIES : return "properties";
		default:
			Debug_assert(DEBUG_ALL, 0, "Type unknown = %d\n", type);
			return NULL;
	}
}


// Properties struct definition

struct Properties {
	int keyMaxLen;
	int stringValueMaxLen;
	HashTable* hTable;  // char[keyMaxLen] --> Properties_Content*
	char* auxKeyMaxLen;
	char* auxValueMaxLen;
};


typedef union {
	double d;
	int i;
	long l;
	unsigned long ul;
	char* s;
	Properties* p;
} Properties_Content_Value;

// Properties_Content struct definition and methods
struct Properties_Content {
	Properties_Content_Type type;
	Properties_Content_Value value;
};

void _Properties_Content_free(Properties_Content** thisP) {
	Properties_Content* this = *thisP;
	if (this->type == STRING) {
		free( (char*)this->value.s );
	}
	else if (this->type == PROPERTIES) {
		Properties* nested = (Properties*)this->value.p;
		Properties_free( &nested );
	}	
	free(this);
	*thisP = NULL;
}


int _fprint_INDENT_STRING(FILE* f, int level) {
	int result = 0;
	for (int i=0; i<level; i++) {
		result += fprintf(f, "%s", INDENT_STRING);
	}
	return result;
}



/** Builds a new Properties_Content instance having type t and a *copy* of *v as value */
Properties_Content* _Properties_Content_new(Properties_Content_Type t, const void* v) {
	Properties_Content* new = calloc(1, sizeof(Properties_Content));
	Debug_assert(DEBUG_P, new != NULL, "new == NULL\n");
	//fprintf(stderr, "start _Properties_Content_new(%p)\n", v);	
	//if (new == NULL), fprintf(stderr, "new == NULL\n");
	new->type = t;
	switch (t) {
		case INT: 
		case BOOLEAN:
			new->value.i = *(int*)v;
			//fprintf(stderr, "new->value.i = %d\n", new->value.i);
			break;
		case ULONG:
			new->value.ul = *(unsigned long*)v;
			break;
		case LONG:
			new->value.l = *(long*)v;
			break;
		case DOUBLE:
			new->value.d = *(double*)v;
			break;
		case STRING:
			new->value.s = StringUtils_clone((char*)v);
			break;
		case PROPERTIES:
			new->value.p = Properties_clone((Properties*)v);
			break;
		default:
			Debug_assert(DEBUG_ALL, 0, "Type unknown = %d\n", t);
			_Properties_Content_free(&new);
			return NULL;
	}
	//fprintf(stderr, "_Properties_Content_new(%p) done\n", v);
	return new;
}
Properties_Content* _Properties_Content_clone(Properties_Content* this) {
	Properties_Content* new = calloc(1, sizeof(Properties_Content));
	new->type = this->type;
	new->value = this->value;
	if (new->type == STRING) {
		new->value.s = StringUtils_clone(this->value.s);
	}
	else if (new->type == PROPERTIES) {
		new->value.p = Properties_clone(this->value.p);
	}
	
	return new;
}

/** Appends the string representation of the value of 'this' to stream 'file'.
	Return 0 on success and an error code (int < 0) on failure
*/	
int Properties_Content_Value_fprint_withIndent(const Properties_Content* const this, FILE* file, int level) {
	int result = 0;
	// fprintf returns the nb of chars written, or an int < 0 in case of error
	switch (this->type) {
		case INT :
			result = fprintf(file, "%d", this->value.i);
			break;		
		case ULONG :
			result = fprintf(file, "%lu", this->value.ul);
			break;		
		case LONG :
			result = fprintf(file, "%ld", this->value.l);
			break;		
		case BOOLEAN :
			result = fprintf(file, "%s", (this->value.i ? "TRUE" : "FALSE" )  );
			break;
		case DOUBLE :
			result = fprintf(file, "%lf", this->value.d);
			break;
		case STRING :
			result = fprintf(file, "%s", this->value.s);
			break;
		case PROPERTIES :
			result = fprintf(file, "{\n");
			result += Properties_fprint_withIndent(this->value.p, file, level+1);
			result += _fprint_INDENT_STRING(file, level);
			result += fprintf(file, "}\n");
			break;
		default:
			Debug_assert(DEBUG_ALL, 0, "Type unknown = %d\n", this->type);
			fprintf(file, "ERROR");
			result = -1;
	}
	return result;
}
int Properties_Content_Value_fprint(const Properties_Content* const this, FILE* file) {
	return Properties_Content_Value_fprint_withIndent(this, file, 0);
}	


// Builds a new Properties_Content instance having type t (!= PROPERTIES) and a *copy* of *value_string as a value (converted to type t)
Properties_Content* _Properties_Content_newFromString(Properties_Content_Type type, const char* const value_string) {
	Debug_assert(DEBUG_ALL, type != PROPERTIES, "type == PROPERTIES\n");
	Debug_out(DEBUG_P, "Parsing %s value '%s'...\n", Properties_Content_Type_toString(type), value_string);
	const void* value = NULL;
	int err = 0;
	Properties_Content_Value value_cv = {0};
	switch (type) {
		case INT :
			err = StringUtils_toInt(value_string, &value_cv.i);
			value = &value_cv.i;
			break;
		case ULONG :
			err = StringUtils_toULong(value_string, &value_cv.ul);
			value = &value_cv.ul;
			break;			
		case LONG :
			err = StringUtils_toLong(value_string, &value_cv.l);
			value = &value_cv.l;
			break;			
		case BOOLEAN :			
			err = StringUtils_toBoolean(value_string, &value_cv.i);
			value = &value_cv.i;
			Debug_out(DEBUG_P, " --> Boolean value parsed: '%s' --> %d (err = %d)\n", 
				value_string, value_cv.i, err);
			break;
		case DOUBLE :
			err = StringUtils_toDouble(value_string, &value_cv.d);
			value = &value_cv.d;
			break;
		case STRING :
			value = value_string;
			break;
		default:
			Debug_assert(DEBUG_ALL, 0, "Type unknown = %d\n", type);
			return NULL;
	}
	if (err) {
		Debug_assert(DEBUG_ALL, 0, "Value '%s' is not of type '%s'\n", 
			value_string, Properties_Content_Type_toString(type));
		return NULL;
	}
	Properties_Content* result = _Properties_Content_new(type, value);
	Debug_out(DEBUG_P, "  --> done. Value is: ");
	Debug_perform(DEBUG_P, Properties_Content_Value_fprint(result, stderr));
	Debug_perform(DEBUG_P, fprintf(stderr, "\n"));
	return result;
}
Properties_Content_Type Properties_Content_type(const Properties_Content* const this) {
	return this->type;
}
const void* const Properties_Content_value(const Properties_Content* const this) {
	const void* value = NULL;
	switch (this->type) {
		case INT: 
			value = &this->value.i;
			break;
		case ULONG:
			value = &this->value.ul;
			break;
		case LONG:
			value = &this->value.l;
			break;
		case BOOLEAN:
			value = &this->value.i;
			break;
		case DOUBLE:
			value = &this->value.d;
			break;
		case STRING:
			value = this->value.s;
			break;
		case PROPERTIES:
			value = this->value.p;
			break;
		default:
			Debug_assert(DEBUG_ALL, 0, "Type unknown = %d\n", this->type);
			return NULL;
	}
	return value;
}




// Properties methods



char* _Properties_fullLengthKey(const Properties* const this, const char* const property) {
	strcpy(this->auxKeyMaxLen, property);
	Debug_assert(DEBUG_ALWAYS, strlen(property) <= this->keyMaxLen, "!!!\n");
	StringUtils_rightPad(this->auxKeyMaxLen, this->keyMaxLen, '\0');
	return this->auxKeyMaxLen;
}


Properties* Properties_new(int keyMaxLen, int stringValueMaxLen) {
	Properties* new = calloc(1, sizeof(Properties));
	Debug_assert(DEBUG_ALL, new != NULL, "new == NULL");	
	Debug_assert(DEBUG_ALL, keyMaxLen > 0, "keyMaxLen = %d <= 0\n", keyMaxLen);
	new->keyMaxLen = keyMaxLen;
	new->stringValueMaxLen = stringValueMaxLen;
	new->hTable = HashTable_new(sizeof(char)*(keyMaxLen+1));
	new->auxKeyMaxLen = calloc(keyMaxLen+1, sizeof(char));
	new->auxValueMaxLen = calloc(stringValueMaxLen+1, sizeof(char));
	return new;
}

Properties* Properties_clone(const Properties* const this) {
	Properties* new = Properties_new(this->keyMaxLen, this->stringValueMaxLen);
	
	HashTableIterator* it = HashTableIterator_new(this->hTable);
	HashTable_Entry* e = NULL;
	while ((e = HashTableIterator_next(it)) != NULL) {
		const char* key = (const char*)HashTable_Entry_key(e);
		Properties_Content* content = (Properties_Content*)HashTable_Entry_value(e);
		HashTable_put(new->hTable, key, _Properties_Content_clone(content));
	}
	HashTableIterator_free(&it);		
	return new;
}


void Properties_clear(Properties* this) {
	HashTableIterator* it = HashTableIterator_new(this->hTable);
	HashTable_Entry* e = NULL;
	while ((e = HashTableIterator_next(it)) != NULL) {
		Properties_Content* content = (Properties_Content*)HashTable_Entry_value(e);
		if (content != NULL) {
			_Properties_Content_free(&content);
		}
		HashTable_clear(this->hTable, (char*)HashTable_Entry_key(e));
	}
	HashTableIterator_free(&it);		
}



void _Properties_set(const Properties* const this, const char* const property, Properties_Content* content) {
	Debug_assert(DEBUG_P, content != NULL, "content == NULL\n");
	char* fullLengthKey = _Properties_fullLengthKey(this, property);
	Properties_Content* old_content = 
		(Properties_Content*)HashTable_put(this->hTable, fullLengthKey, content);

	if (old_content != NULL) {
		Debug_out(DEBUG_P, "Value for property '%s' redefined. New value:\n", property);
		Debug_perform(DEBUG_P, Properties_Content_Value_fprint(content, stderr));
		// free old value
		_Properties_Content_free(&old_content);
	}
}

void Properties_setDouble(const Properties* const this, const char* const property, double value) {
	_Properties_set(this, property, 
		_Properties_Content_new(DOUBLE, &value));
}
void Properties_setInt(const Properties* const this, const char* const property, int value) {
	_Properties_set(this, property, 
		_Properties_Content_new(INT, &value));
}
void Properties_setULong(const Properties* const this, const char* const property, unsigned long value) {
	_Properties_set(this, property, 
		_Properties_Content_new(ULONG, &value));
}
void Properties_setLong(const Properties* const this, const char* const property, long value) {
	_Properties_set(this, property, 
		_Properties_Content_new(LONG, &value));
}
void Properties_setBoolean(const Properties* const this, const char* const property, int value) {
	//Debug_out(DEBUG_P, "setBoolean()\n");	
	//fprintf(stderr, "start setBoolean()\n");	
	if (this == NULL) fprintf(stderr, "this==NULL\n");
	Debug_assert(DEBUG_P, this != NULL, "this == NULL\n");
	//if (property == NULL) fprintf(stderr, "property==NULL\n");	
	Debug_assert(DEBUG_P, property != NULL, "property == NULL\n");	
	//Debug_out(DEBUG_P, "creating content\n");
	//fprintf(stderr, "creating content from &value = %p\n", (void*)&value);
	Properties_Content* content = _Properties_Content_new(BOOLEAN, &value);
	//Debug_assert(DEBUG_P, content != NULL, "content == NULL\n");		
	//if (content == NULL) fprintf(stderr, "content==NULL\n");
	//fprintf(stderr, "content is %p\n", (void*)&content);
	_Properties_set(this, property, content);
	//fprintf(stderr, "setBoolean() done\n");
}
void Properties_setString(const Properties* const this, const char* const property, const char* const value) {
	_Properties_set(this, property, 
		_Properties_Content_new(STRING, value));
}
void Properties_setNestedProperties(const Properties* const this, const char* const property, const Properties* value) {
	_Properties_set(this, property, 
		_Properties_Content_new(PROPERTIES, value));
}

// Setter methods: the value is always cloned (also if it is of type Properties)
void Properties_set(const Properties* const this, const char* const property, Properties_Content_Type type, const void* value) {
	_Properties_set(this, property, _Properties_Content_new(type, value));
}

int Properties_setFromString(const Properties* const this, const char* const property, Properties_Content_Type type, const char* const value_string) {
	Properties_Content* content = _Properties_Content_newFromString(type, value_string);
	if (content == NULL) return -1;
	_Properties_set(this, property, content);
	return 0;
}


// File syntax:  <type> <property> = <value>
Properties* Properties_newFromFile(const char* const fname, int keyMaxLen, int stringValueMaxLen) {
	Debug_out(DEBUG_P, "Properties_newFromFile(): start\n");
	
	Array* nestedPropertiesStack = Array_new(1024, sizeof(Properties*), 1024, NULL, NULL);
	Array* nestedPropertiesNamesStack = Array_new(1024, sizeof(char*), 1024, NULL, NULL);	
	
	Properties* new = Properties_new(keyMaxLen, stringValueMaxLen);
	Array_push(nestedPropertiesStack, &new);
	Debug_out(DEBUG_P, "Properties_newFromFile(): pushing Properties (new): %p\n", (void*)new);
	
	FILE* f = fopen(fname, "r");
	Debug_assert(DEBUG_ALL, f != NULL, "Properties_newFromFile(): could not open file '%s'\n", fname);
	if (f == NULL) {
		perror("Could not open Properties file\n");
		return NULL;
	}
	char* line = calloc(LINE_MAX_LEN+1, sizeof(char));
	int lnumber = 0;
	Array* pv_tokens = Array_newString(2, 1);  // (type + property) and value
	Array* tp_tokens = Array_newString(2, 1); // type and property
	char* s = NULL;	
	
	// Read file lines
	while (fgets(line, LINE_MAX_LEN, f) != NULL) {		
		lnumber++;		
		Debug_out(DEBUG_P, "newFromFile(): reading line %d\n", lnumber);
		int startComment = StringUtils_strpos(line, '#');
		if (startComment >= 0) {
			Debug_out(DEBUG_P, "line %d: found comment sign at column %d\n", lnumber, startComment);
			line[startComment] = '\0';
		}		
		StringUtils_trim(line);
		Debug_out(DEBUG_P, "line %d: '%s'\n", lnumber, line);
		// Skip comments
		if (strlen(line) == 0) continue;

		// check whether end-of-properties block is reached (i.e., line is "}")
		if (StringUtils_equals(line, "}")) {
			Properties* nested = NULL;
			Array_pop(nestedPropertiesStack, &nested);
			char* property = NULL;
			Array_pop(nestedPropertiesNamesStack, &property);			
			Debug_assert(DEBUG_ALL, property != NULL, "property == NULL\n");
			Debug_out(DEBUG_P, "Properties_newFromFile(): popping Properties '%s': %p\n",
				property, (void*)nested);
			Properties* currentProperties = NULL;
			Array_top(nestedPropertiesStack, &currentProperties);
			Properties_Content* content =
					_Properties_Content_new(PROPERTIES, nested);
			if (content == NULL) {
				Debug_assert(DEBUG_ALL, content != NULL, "content == NULL\n");
				fprintf(stderr, "Error in file '%s': could not parse correctly value for property name "
					"'%s' of type Properties\n", 
					fname, property);
			}
			_Properties_set(currentProperties, property, content);
			Properties_free(&nested);
			free(property);
		}
		else {
		// Tokenize line: type propertyName = propertyValue
			Array_clear(pv_tokens);
			StringUtils_tokenize(line, "=", pv_tokens);
			if (Array_length(pv_tokens) != 2) {
				fprintf(stderr, "Error in reading Properties from file '%s':%d. "
					"I've read %d tokens separated by '=', 2 expected\n", 
					fname, lnumber, Array_length(pv_tokens));
				Debug_assert(DEBUG_ALL, 0, "Error in reading Properties from file '%s':%d. "
					"I've read %d tokens, 2 expected\n", 
					fname, lnumber, Array_length(pv_tokens));
			}
			Debug_out(DEBUG_P, "newFromFile():%d -- A\n", lnumber);
			char* type_and_property = NULL;
			Array_get(pv_tokens, 0, &type_and_property);
			char* value_string = NULL;
			Array_get(pv_tokens, 1, &value_string);

			StringUtils_trim(type_and_property);
			StringUtils_trim(value_string);
		
			Array_clear(tp_tokens);
			StringUtils_tokenize(type_and_property, " ", tp_tokens);
		
			char* type_string = NULL;
			Properties_Content_Type type = STRING;
			char* property = NULL;
			if (Array_length(tp_tokens) == 1) {
				// no explicit type: type is set to :
				// - Properties, if value == '{'
				// - String, otherwise
				if (StringUtils_equals(value_string, "{")) {
					type_string = "properties";
					type = PROPERTIES;
				} else {
					type_string = "string";
					type = STRING;
				}
				Array_get(tp_tokens, 0, &property);				
			} else if (Array_length(tp_tokens) != 2) {
				fprintf(stderr, "Error in reading type and property name from file '%s':%d. "
					"I've read %d tokens, 2 expected\n", 
					fname, lnumber, Array_length(tp_tokens));
				Debug_assert(DEBUG_ALL, 0, "Error in reading type and property name from file '%s':%d. "
					"I've read %d tokens, 2 expected\n", 
					fname, lnumber, Array_length(tp_tokens));
				continue; // if assertions are disabled, ignores this line
			} else {
				// exactly 2 tokens: "type" "property"
				Array_get(tp_tokens, 0, &type_string);		
				Array_get(tp_tokens, 1, &property);
				StringUtils_trim(type_string);
				type = _Properties_Content_Type_parse(type_string);		
			}
			StringUtils_trim(property);
		
			// line read in vars: type "property" = "value_string"
		
			Debug_out(DEBUG_P, "newFromFile(): type = '%s', property = '%s', value = '%s'\n",
				type_string, property, value_string);
		
			if (type == PROPERTIES) { 
				// nested properties start
				Debug_assert(DEBUG_ALL, StringUtils_equals(value_string, "{"), 
					"{ '<newline>' missing after "
					"'Properties pname = '\n");
				Properties* nested = Properties_new(keyMaxLen, stringValueMaxLen);
				Array_push(nestedPropertiesStack, &nested);
				Debug_out(DEBUG_P, "Properties_newFromFile(): pushing Properties '%s': %p\n",
				 	property, (void*)nested);
				char* property_clone = StringUtils_clone(property);
				Array_push(nestedPropertiesNamesStack, &property_clone);
			} else {
				// "single-line" property
				Properties_Content* content =
					_Properties_Content_newFromString(type, value_string);
				if (content == NULL) {
					Debug_assert(DEBUG_ALL, content != NULL, "content == NULL\n");
					fprintf(stderr, "Error in file '%s': value '%s' is not a value of type '%s'. "
						"Skipping...", 
						fname, value_string, type_string);
				}
				Properties* currentProperties = NULL;
				Array_top(nestedPropertiesStack, &currentProperties);
				Debug_assert(DEBUG_ALL, currentProperties != NULL, "currentProperties == NULL\n");
				Debug_out(DEBUG_P, "Adding single-line property '%s' property to Properties %p...\n", 
					property, (void*)currentProperties);
				_Properties_set(currentProperties, property, content);
				Debug_out(DEBUG_P, "  --> done\n");
			}
		}
	}
	
	// check whether nested properties has all been terminated
	Debug_assert(DEBUG_ALL, Array_length(nestedPropertiesStack) == 1, 
		"Error: Missing '}' somewhere in file file '%s'\n", fname);
	
	int err = fclose(f);
	if (err) {
		fprintf(stderr, "Error while closing file '%s'. fclose() returned error %d\n", fname, err);
		Debug_assert(DEBUG_ALL, 0, "Error while closing file '%s'. "
			"fclose() returned error %d\n", fname, err);
	}
	return new;
}



int Properties_fprint(const Properties* const this, FILE* f) {
	return Properties_fprint_withIndent(this, f, 0);
}


int Properties_fprint_withIndent(const Properties* const this, FILE* f, int level) {	
	Properties_Iterator* it = Properties_Iterator_new(this);
	const Properties_Entry* e = NULL;
	int result = 0;
	while ((e = Properties_Iterator_next(it)) != NULL) {
		const char* property = Properties_Entry_property(e);
		const Properties_Content* content = Properties_Entry_value(e);
						
		result += _fprint_INDENT_STRING(f, level);
		
		result += fprintf(f, "%s %s = ", 
			Properties_Content_Type_toString(content->type), 
			property
		);
		result += Properties_Content_Value_fprint_withIndent(content, f, level);
		result += fprintf(f, "\n");
	}	
	Properties_Iterator_free(&it);
	return result;
}

/*
void Properties_toString(const Properties* const this, char* result, int length) {
	Debug_assert(DEBUG_P, this != NULL, "this == NULL\n");
	Debug_assert(DEBUG_P, result != NULL, "result == NULL\n");
	result[0] = '\0';
	char* resultTail = result;
		
	HashTableIterator* it = HashTableIterator_new(this->hTable);
	HashTable_Entry* e = NULL;
	int charsWritten = 0;
		
	char line[length] = {0};
	while ((e = HashTableIterator_next(it)) != NULL) {
		char* property = (char*)HashTable_Entry_key(e);
		Properties_Content* content = (Properties_Content*)HashTable_Entry_value(e);
						
		charsWritten += snprintf(line, length, "%s %s = %s\n", 
			Properties_Content_Type_toString(content->type), 
			property, 
			Properties_Content_Value_toString(content, this->auxValueMaxLen)
		);
		if (charsWritten > length) {
			break;
		}
		length -= charsWritten;		
		strncat(resultTail, line, length);
		resultTail += charsWritten;
	}
	HashTableIterator_free(&it);	
}
*/

int Properties_save(const Properties* const this, char* filePath) {
	FILE* f = fopen(filePath, "w");
	if (f == NULL) {
		fprintf(stderr, "Cannot open file '%s' to write properties\n", filePath);
		return -1;
	}
	Properties_fprint(this, f);
	int err = fclose(f);
	if (err) {
		fprintf(stderr, "Cannot close file '%s' after writing. fclose() returned error %d\n", filePath, err);
		return -1;
	}
	return 0;
}	

unsigned long Properties_size(const Properties* const this) {
	return HashTable_size(this->hTable);
}


// Getter methods
int Properties_contains(const Properties* const this, const char* const property) {
	const Properties_Content* const result = Properties_get(this, property);
	return (result != NULL);
}

const Properties_Content* const Properties_get(const Properties* const this, const char* const property) {
	Properties_Content* result = NULL;
	HashTable_get(this->hTable, 
		_Properties_fullLengthKey(this, property), 
		(void**) &result);
	return result;
}
int Properties_getInt(const Properties* const this, const char* const property, int* value) {	
	const Properties_Content* const result = Properties_get(this, property);
	if (result == NULL) return -1;
	if (result->type != INT) return -2;
	*value = result->value.i;
	return 0;
}
int Properties_getULong(const Properties* const this, const char* const property, unsigned long* value) {	
	const Properties_Content* const result = Properties_get(this, property);
	if (result == NULL) return -1;
	if (result->type != ULONG) return -2;
	*value = result->value.ul;
	return 0;
}
int Properties_getLong(const Properties* const this, const char* const property, long* value) {	
	const Properties_Content* const result = Properties_get(this, property);
	if (result == NULL) return -1;
	if (result->type != LONG) return -2;
	*value = result->value.l;
	return 0;
}
int Properties_getBoolean(const Properties* const this, const char* const property, int* value) {	
	const Properties_Content* const result = Properties_get(this, property);
	if (result == NULL) return -1;	
	if (result->type != BOOLEAN) return -2;
	*value = result->value.i;
	return 0;
}
int Properties_getDouble(const Properties* const this, const char* const property, double* value) {	
	const Properties_Content* const result = Properties_get(this, property);
		if (result == NULL) return -1;
	if (result->type != DOUBLE) return -2;
	*value = result->value.d;
	return 0;
}
int Properties_getString(const Properties* const this, const char* const property, const char** const value) {	
	const Properties_Content* const result = Properties_get(this, property);
	if (result == NULL) return -1;	
	if (result->type != STRING) return -2;
	*value = result->value.s;
	return 0;
}
int Properties_getNestedProperties(const Properties* const this, const char* const property, const Properties**  const value) {
	const Properties_Content* const result = Properties_get(this, property);
	if (result == NULL) return -1;	
	if (result->type != PROPERTIES) return -2;
	*value = result->value.p;
	return 0;
	
}


Properties_Content_Type Properties_type(const Properties* const this, const char* const property) {
	const Properties_Content* const content = Properties_get(this, property);
	Debug_assert(DEBUG_ALL, content != NULL, "Property '%s' not found\n", property);
	return content->type;
}

void Properties_free(Properties** thisP) {
	Debug_assert(DEBUG_ALL, thisP != NULL, "thisP == NULL\n");	
	Properties* this = *thisP;

	HashTableIterator* it = HashTableIterator_new(this->hTable);
	HashTable_Entry* e = NULL;
	while ((e = HashTableIterator_next(it)) != NULL) {
		Properties_Content* content = (Properties_Content*)HashTable_Entry_value(e);
		_Properties_Content_free(&content);
	}
	HashTableIterator_free(&it);
	HashTable_free(&this->hTable);
	free(this->auxKeyMaxLen);
	free(this->auxValueMaxLen);
	free(this);
	*thisP = NULL;
}





// Properties Iterator
struct Properties_Iterator {
	HashTableIterator* hTableIt;
};

/** Create an Iterator for this Properties */
Properties_Iterator* Properties_Iterator_new(const Properties* const this) {
	Properties_Iterator* new = calloc(1, sizeof(Properties_Iterator));
	Debug_assert(DEBUG_P, new != NULL, "new == NULL\n");
	new->hTableIt = HashTableIterator_new(this->hTable);
	Debug_assert(DEBUG_P, new->hTableIt != NULL, "new->hTableIt == NULL\n");	
	return new;
}

void Properties_Iterator_rewind(Properties_Iterator* const this) {
	HashTableIterator_rewind(this->hTableIt);
}

/** Return 1 if a call to Properties_Iterator_next(this) will return an entry, 0 otherwise */
int Properties_Iterator_hasNext(const Properties_Iterator* const this) {
	return HashTableIterator_hasNext(this->hTableIt);
}

/** Return the next entry of the HashTable which 'this' is iterating on, 
	NULL if no more entries exist */
const Properties_Entry* const Properties_Iterator_next(Properties_Iterator* const this) {
	return HashTableIterator_next(this->hTableIt);
}

const char* const Properties_Entry_property(const Properties_Entry* const this) {
	return (char*)HashTable_Entry_key(this);
}
const char* const Properties_Entry_key(const Properties_Entry* const this) {
	return Properties_Entry_property(this);
}
const Properties_Content* const Properties_Entry_value(const Properties_Entry* const this) {
	return (Properties_Content*)HashTable_Entry_value(this);
}

void Properties_Iterator_free(Properties_Iterator** thisP) {
	Debug_assert(DEBUG_P, thisP != NULL, "thisP == NULL\n");
	Properties_Iterator* this = *thisP;
	HashTableIterator_free(&this->hTableIt);
	free(this);
	*thisP = NULL;
}



