#include <stdio.h>
#include <stdlib.h>

#include <Debug.h>

#include <Range_Int.h>

#define DEBUG "Range_Int"



struct Range_Int {
	int min;
	int max;
	int step;
	int greatest; // greatest value
	
	// cache
	int size;
};

struct Range_Int_Iterator {
	const Range_Int* range;
	
	int next;
	int prev;
	int step;
	int looping;
	
};

int _size_debug(const Range_Int* const this) {
	int result = 0;
	for(int i = this->min; i <= this->max; i += this->step) {
		result++;
	}
	return result;
}



Range_Int* Range_Int_new(int min, int max, int step) {
	Debug_assert(DEBUG_ALWAYS, max >= min, "max = %d < %d = min\n", max, min);
	Debug_assert(DEBUG_ALWAYS, step > 0, "step = %d <= '\n", step);
	Range_Int* new = calloc(1, sizeof(Range_Int));
	Debug_assert(DEBUG_ALWAYS, new != NULL, "new == NULL\n");
	new->min = min;
	new->max = max;
	new->step = step;
	
	new->greatest = min + step*((max-min)/step);
	Debug_assert(DEBUG_ALWAYS, new->greatest >= min, "new->greatest = %d < min = %d\n",
		new->greatest, min);
	Debug_assert(DEBUG_ALWAYS, new->greatest <= max, "new->greatest = %d > max = %d\n",
		new->greatest, max);
	Debug_assert(DEBUG_ALWAYS, new->greatest+step > max, "new->greatest + step = "
		"%d + %d <= max = %d\n", 
		new->greatest, step, max);
		
	new->size = 1 + (max-min)/step;
	Debug_assert(DEBUG, new->size == _size_debug(new), "new->size = %d != "
		"%d = _size_debug()\n", new->size, _size_debug(new));
	return new;
}

int Range_Int_size(const Range_Int* const this) {
	Debug_assert(DEBUG_ALWAYS, this != NULL, "this == NULL\n");
	return this->size;
}



int Range_Int_min(const Range_Int* const this) {
	Debug_assert(DEBUG_ALWAYS, this != NULL, "this == NULL\n");
	return this->min;
}
int Range_Int_max(const Range_Int* const this) {
	Debug_assert(DEBUG_ALWAYS, this != NULL, "this == NULL\n");	
	return this->max;
}
int Range_Int_step(const Range_Int* const this) {
	Debug_assert(DEBUG_ALWAYS, this != NULL, "this == NULL\n");	
	return this->step;
}
void Range_Int_free(Range_Int** thisP) {
	Debug_assert(DEBUG_ALWAYS, thisP != NULL, "thisP == NULL\n");
	Range_Int* this = *thisP;
	if (this == NULL) return;
	free(this);
	*thisP = NULL;
}


int Range_Int_index(const Range_Int* const this, int value, int* indexP) {
	Debug_assert(DEBUG, this != NULL, "this == NULL\n");
	Debug_assert(DEBUG, indexP != NULL, "indexP == NULL\n");	
	int i = ( (value-this->min) / this->step);
	if (i * this->step == (value - this->min)) {
		*indexP = i;
		return 0;
	} else {
		return -1;
	}
}
int Range_Int_value(const Range_Int* const this, int index, int* valueP) {
	Debug_assert(DEBUG, this != NULL, "this == NULL\n");
	Debug_assert(DEBUG, valueP != NULL, "valueP == NULL\n");
	if (index < 0 || index >= Range_Int_size(this)) return -1;
	
	int result = this->min + this->step * index;
	if (result < this->min || result > this->max) {
		return -1;
	}
	*valueP = result;
	return 0;
}


int Range_Int_contains(const Range_Int* const this, int value) {
	int diff = value - this->min;
	return (diff % this->step == 0);
}

Range_Int_Iterator* Range_Int_Iterator_new(const Range_Int* const range) {
	Debug_assert(DEBUG_ALWAYS, range != NULL, "range == NULL");
	Range_Int_Iterator* new = calloc(1, sizeof(Range_Int_Iterator));
	Debug_assert(DEBUG_ALWAYS, new != NULL, "new == NULL");
	new->range = range;
	new->step = range->step;
	new->looping = 0;
	Range_Int_Iterator_rewind(new);
	return new;
}
Range_Int_Iterator* Range_Int_Iterator_new_looping(const Range_Int* const range) {
	Range_Int_Iterator* new = Range_Int_Iterator_new(range);
	new->looping = 1;
	return new;
}

void Range_Int_Iterator_setStep(Range_Int_Iterator* const this, int step) {
	Debug_assert(DEBUG_ALWAYS, this != NULL, "this == NULL\n");		
	Debug_assert(DEBUG_ALWAYS, step > 0, "step = %d <= 0\n", step);
	this->step = step;
}

int Range_Int_Iterator_hasNext(Range_Int_Iterator* const this) {
	Debug_assert(DEBUG_ALWAYS, this != NULL, "this == NULL\n");	
	return this->looping || this->next <= this->range->max;
}	
int Range_Int_Iterator_hasPrev(Range_Int_Iterator* const this) {
	Debug_assert(DEBUG_ALWAYS, this != NULL, "this == NULL\n");	
	return this->looping || this->prev >= this->range->min;
}	


void Range_Int_Iterator_fprint(const Range_Int_Iterator* const this, FILE* f) {
	fprintf(f, "[prev=%d, next=%d] ", this->prev, this->next);
}

int Range_Int_Iterator_next(Range_Int_Iterator* const this, int* result) {
	if (!Range_Int_Iterator_hasNext(this)) return -1;
	Debug_assert(DEBUG_ALWAYS, result != NULL, "result == NULL\n");

	if (this->looping) {
		if (this->next > this->range->max) {
			this->next = (this->range->min + (this->next - this->range->max) - 1);
		}
	}	
	*result = this->next;
	this->prev = *result - this->step;
	this->next += this->step;
	return 0;
}
int Range_Int_Iterator_prev(Range_Int_Iterator* const this, int* result) {
	if (!Range_Int_Iterator_hasPrev(this)) return -1;
	Debug_assert(DEBUG_ALWAYS, result != NULL, "result == NULL\n");
	if (this->looping) {	
		if (this->prev < this->range->min) {
			this->prev = (1 + this->range->max - (this->range->min - this->prev));
		}
	}	
	*result = this->prev;
	this->next = *result + this->step;
	this->prev -= this->step;
	return 0;
}

void Range_Int_Iterator_rewind(Range_Int_Iterator* const this) {
	Debug_assert(DEBUG_ALWAYS, this != NULL, "this == NULL\n");	
	this->next = this->range->min;
	this->prev = this->range->greatest;
}
void Range_Int_Iterator_free(Range_Int_Iterator** thisP) {
	Debug_assert(DEBUG_ALWAYS, thisP != NULL, "thisP == NULL\n");
	Range_Int_Iterator* this = *thisP;
	if (this == NULL) return;
	free(this);
	*thisP = NULL;
}



