#include <Server.h>
#include <Debug.h>
#include <StringUtils.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <strings.h>


#define DEBUG "Server"




static int _Server_openSocketAtPort(Server *_this);
static void * _Server_get_in_addr(struct sockaddr *sa);



struct Server {
	const char * address;
	int port;

	const char * name;
	int maxMsgSize;
	int maxAcceptedConnAtTime;
	const char * controlStopServer;
};

Server * Server_new(const char *name, const char * address, int port, const char *controlStopServer, int maxMsgSize, int maxAccConnAtTime)
{
	Debug_assert(DEBUG, name != NULL, "name == NULL\n");
	Server * new = (Server *) calloc(1, sizeof(Server));
	new->name = StringUtils_clone(name);
	if (address != NULL) {
		new->address = StringUtils_clone(address);
	}
	new->port = port;
	if (controlStopServer != NULL) {
		new->controlStopServer = StringUtils_clone(controlStopServer);
	}
	new->maxMsgSize = maxMsgSize;
	new->maxAcceptedConnAtTime = maxAccConnAtTime;
	return new;
}

void Server_free(Server **_this)
{
	Debug_assert(DEBUG, *_this != NULL, "*_this == NULL\n");
	free(*_this);
	*_this = NULL;
}


/**
 *	Accepts one connection and everything is done in between
 */
pid_t Server_start(Server *_this, char * (*parsingFnc)(void *, char *), void *parsingFncFirstPar, FILE *stream)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	pid_t childProcess;

	childProcess = fork();

	/* Child code */
	if (childProcess == 0) {
		/* child is server always accepting requests */
		int fncres;
		struct sockaddr_storage their_addr; // connector's address information
		socklen_t sin_size;
		pid_t server_pid = getpid(); // pid of server
		int new_fd;
		char s[INET6_ADDRSTRLEN];

		/* Open server's socket */
		int sock = _Server_openSocketAtPort(_this);
		if (sock == -1) {
			exit(-1);
		}
		Debug_out(DEBUG, "[%s] Opened server at port %d\n", _this->name, _this->port);

		/* accept and close when control signal STOP is read */
		Debug_out(DEBUG, "[%s] Accepting request on %d\n", _this->name, _this->port);


		sin_size = sizeof(their_addr);

		new_fd = accept(sock, (struct sockaddr *)&their_addr, &sin_size);

		if (new_fd == -1) {
			Debug_out(DEBUG, "[%s] ERROR accepting\n", _this->name);
			perror("");
			fflush(NULL);
			exit(-1);
		}

		inet_ntop(their_addr.ss_family,
				  _Server_get_in_addr((struct sockaddr *)&their_addr),
				  s, sizeof(s));
		Debug_out(DEBUG, "[%s] Connection from %s\n", _this->name, s);

		while (1) {
			static int numbytes;
			static const char * parsingFncRes;
			char * buf = calloc(_this->maxMsgSize, sizeof(char));

			/* reads a message */
			numbytes = recv(new_fd, buf, _this->maxMsgSize, 0);
			if (numbytes == -1) {
				Debug_out(DEBUG, "[%s] ERROR receiving a message\n", _this->name);
				perror("");
				fflush(NULL);
				continue;
			}
			else if (numbytes == 0) { // client has closed connection
				Debug_out(DEBUG, "[%s] Client has closed its connection\n", _this->name);
				break;
			}
			buf[numbytes] = '\0';
			Debug_out(DEBUG, "[%s] Received message '%s'\n", _this->name, buf);

			/* check if it is stop command */
			if (_this->controlStopServer && StringUtils_equals(_this->controlStopServer, buf)) {
				close(new_fd);  // parent doesn't need this descriptor
				break;
			}

			/* do something with this message */
			parsingFncRes = parsingFnc(parsingFncFirstPar, buf);
			Debug_out(DEBUG, "parsingFncRes = '%s'\n", parsingFncRes? parsingFncRes : "NULL");

			if (!parsingFncRes) {
				Debug_out(DEBUG, "[%s] Error parsing message '%s'\n", _this->name, buf);
			} else {
				/* possibly answer to request */
				if (!StringUtils_equals(parsingFncRes, "")) {
					numbytes = send(new_fd, parsingFncRes, strlen(parsingFncRes), 0);
					if (numbytes <= 0) {
						Debug_out(DEBUG, "[%s] ERROR sending message %s\n", _this->name, parsingFncRes);
						perror("");
						fflush(NULL);
						continue;
					}
					Debug_out(DEBUG, "Sent message '%s'\n", parsingFncRes);
				}
			}
			free(buf);
		}


		/* server shuts down */
		close(new_fd);
		close(sock);
		Debug_out(DEBUG, "[%s] Closed server at port %d\n", _this->name, _this->port);
		exit(0);
	}
	else if (childProcess < 0) { /* Error in fork() */
		Debug_out(DEBUG, "[%s] Error in creating server\n", _this->name);
		perror("");
		fflush(NULL);
		return -1;
	}

	return childProcess;
}


void Server_client(Server *_this)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
}


int Server_send(Server *_this, char *msg)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	return 0;
}



int Server_receive(Server *_this, char *msg)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	return 0;
}











/* Returns socket fd if everything is OK, -1 otherwise */
static int _Server_openSocketAtPort(Server *_this)
{
	Debug_assert(DEBUG, _this != NULL, "_this == NULL\n");
	struct sockaddr_in sin;

	/* Creates socket */
	int sock = socket(AF_INET, SOCK_STREAM, 0);
	if(sock == -1) {
		fprintf(stderr, "[%s] Error opening socket\n", _this->name);
		perror("");
		return -1;
	}
	sin.sin_port = htons(_this->port);
	sin.sin_addr.s_addr = 0;
	sin.sin_addr.s_addr = INADDR_ANY;
	sin.sin_family = AF_INET;
	/* Binds socket */
	if(bind(sock, (struct sockaddr *)&sin,sizeof(struct sockaddr_in) ) == -1) {
		fprintf(stderr, "[%s] Error binding socket at port %d\n", _this->name, _this->port);
		perror("");
		return -1;
	}
	/* Listens on socket */
	if (listen(sock, _this->maxAcceptedConnAtTime) == -1) {
		fprintf(stderr, "[%s] Error listening on soket at port %d\n", _this->name, _this->port);
		perror("");
		return -1;
	}

	return sock;
}



// get sockaddr, IPv4 or IPv6:
void * _Server_get_in_addr(struct sockaddr *sa)
{
	if (sa->sa_family == AF_INET) {
		return &(((struct sockaddr_in*)sa)->sin_addr);
	}

	return &(((struct sockaddr_in6*)sa)->sin6_addr);
}

